"""
Tracks changes between deployments.
"""
from __future__ import print_function

#TODO: remove? largely deprecated, use the deploy module instead

from fabric.api import (
    env,
    runs_once,
)

from burlap import common
from burlap.decorators import task_or_dryrun

@task_or_dryrun
@runs_once
def show(name):
    name = name.strip().lower()
    func = common.manifest_recorder[name]
    ret = func()
    print(ret)

@task_or_dryrun
@runs_once
def get_current(name):
    name = name.strip().lower()
    func = common.manifest_recorder[name]
    return func()
    
@task_or_dryrun
@runs_once
def get_last(name):
    from burlap.deploy import get_last_thumbprint
    name = common.assert_valid_satchel(name)
    last_thumbprint = get_last_thumbprint()
    if last_thumbprint:
        if name in last_thumbprint:
            return last_thumbprint.get(name, type(env)())
    return type(env)()
    
@task_or_dryrun
@runs_once
def changed(name):
    from burlap.deploy import get_last_thumbprint
    name = name.strip().lower()
    if name not in common.manifest_recorder:
        print('No manifest recorder has been registered for component "%s"' % name)
    else:
        last_thumbprint = get_last_thumbprint()
        if last_thumbprint:
            if name in last_thumbprint:
                last_manifest = last_thumbprint[name]
                current_manifest = common.manifest_recorder[name]()
                if last_manifest == current_manifest:
                    print('No')
                    return False
                else:
                    print('Yes')
                    return True
            else:
                print('Yes, first deployment for this component.')
                return True
        else:
            print('Yes, first deployment.')
            return True



from __future__ import print_function

import os
import sys
import shutil

from fabric.api import env

from burlap.shelf import Shelf
from burlap.context import set_cwd
from burlap.common import getoutput, set_verbose
from burlap.deploy import preview as deploy_preview
from burlap import load_role_handler
#from burlap.vagrant import vagrant

def test_deploy():
    """
    Creates a multi-site Apache Django powered web server with a MySQL backend.
    """
    
    try:
        set_verbose(True)
        
        # Delete any old tmp files
        PROJECT_DIR = '/tmp/burlap_test'
        if os.path.exists(PROJECT_DIR):
            shutil.rmtree(PROJECT_DIR)
        os.makedirs(PROJECT_DIR)
     
        # Create our test virtualenv.
        PYTHON_EXE = os.path.split(sys.executable)[-1]
        VIRTUALENV_DIR = os.path.join(PROJECT_DIR, '.env')
        BURLAP_DIR = os.path.abspath('./burlap')
        BURLAP_BIN = os.path.abspath('./bin/burlap-admin.py')
        SITE_PACKAGES = os.path.join(VIRTUALENV_DIR, 'lib/%s/site-packages' % PYTHON_EXE)
     
        # Initialize project.
        kwargs = dict(
            project_dir=PROJECT_DIR,
            burlap_bin=BURLAP_BIN,
        )
        print('Initializing project skeleton...')
        cmd = 'cd {project_dir}; {burlap_bin} skel multitenant'.format(**kwargs)
        print('cmd:', cmd)
        assert not os.system(cmd)
     
        # Symlink burlap.
        assert not os.system('ln -s %s %s' % (BURLAP_DIR, SITE_PACKAGES))
        
        # Add production role.
        VIRTUALENV_ACTIVATE = '. %s/bin/activate' % VIRTUALENV_DIR
        kwargs = dict(
            project_dir=PROJECT_DIR,
            activate=VIRTUALENV_ACTIVATE,
            burlap_bin=BURLAP_BIN,
        )
        assert not os.system('cd {project_dir}; {burlap_bin} add-role prod'.format(**kwargs))
         
        # Test logging in to VM.
        print('env.host_string:', env.host_string, env.host)
        print('env.user:', env.user)
        print('env.key_filename:', env.key_filename)
        env.ROLE = 'prod'
        prod_settings = Shelf(filename=PROJECT_DIR+'/roles/{role}/settings.yaml')
        prod_settings['hosts'] = [env.host] if isinstance(env.host, basestring) else env.host
        #prod_settings['host_string']
        prod_settings['user'] = env.user
        prod_settings['key_filename'] = env.key_filename
        prod_settings['is_local'] = False
        prod_settings['app_name'] = 'multitenant'
        kwargs = dict(
            project_dir=PROJECT_DIR,
        )
        print('Testing hello world...')
        cmd = 'cd {project_dir}; . ./setup.bash; fab prod:verbose=1 shell:command="echo hello"'.format(**kwargs)
        print('cmd:', cmd)
        assert not os.system(cmd)
        print('Testing ifconfig...')
        cmd = 'cd {project_dir}; . ./setup.bash; fab prod:verbose=1 shell:command="ifconfig"'.format(**kwargs)
        print('cmd:', cmd)
        out = getoutput(cmd)
        print('out:', out)
        assert 'inet addr:127.0.0.1' in out
        
        # Add services.
        services = prod_settings.get('services', [])
        services.extend([
            'apache',
            #'hostname',
            'mysql',
            'mysqlclient',
            'ntpclient',
            'packager',
            'pip',
            'sshnice',
            'tarball',
            'timezone',
            'ubuntumultiverse',
            'unattendedupgrades',
            #'user',
        ])
        prod_settings.set('services', services)
        prod_settings.set('sites', {
            'multitenant': {
                'apache_domain_template': 'multitenant.test.com',
                'apache_domain_with_sub_template': 'multitenant.test.com',
                'apache_domain_without_sub_template': 'multitenant.test.com',
                'apache_server_aliases_template': 'multitenant.test.com',
                'apache_ssl': False,
                'apache_auth_basic': False,
                'apache_enforce_subdomain': False,
            }
        })
        prod_settings.set('pip_requirements', 'pip-requirements.txt')
        
        # Confirm deployment changes are detected.
        with set_cwd(PROJECT_DIR):
            #from burlap import role_prod as prod
            prod = load_role_handler('prod')
            prod()
            assert 'app_name' in env
            assert 'sites' in env
            env.host_string = env.hosts[0]
            
            
            changed_components, deploy_funcs = deploy_preview()
            changed_components = sorted(changed_components)
            print('changed_components:', changed_components)
            assert changed_components == [
                'APACHE',
                'MYSQL',
                'MYSQLCLIENT',
                'NTPCLIENT',
                'PACKAGER',
                'PIP',
                'SSHNICE',
                'TIMEZONE',
                'UBUNTUMULTIVERSE',
                'UNATTENDEDUPGRADES',
            ]
            deploy_funcs = sorted(deploy_funcs)
            print('deploy_funcs:', deploy_funcs)
            assert deploy_funcs == [
                ('apache.configure', None),
                ('mysql.configure', None),
                ('mysqlclient.configure', None),
                ('ntpclient.configure', None),
                ('packager.configure', None),
                ('pip.configure', None),
                ('sshnice.configure', None),
                ('timezone.configure', None),
                ('ubuntumultiverse.configure', None),
                ('unattendedupgrades.configure', None),
            ]
        
        # Deploy changes.
        
        # Confirm changes have been cleared.
        
        # Add Django site.
        
    finally:
        # Undo changes to the VM.
        pass
        


from __future__ import print_function

import os
from pipes import quote
# import posixpath
import random
import string

# from fabric.api import (
#     env,
#     settings,
#     task,
# )
from fabric.api import hide

from burlap.constants import *
from burlap import Satchel
# from burlap.common import (
#     QueuedCommand,
# )
# from burlap.group import (
#     exists as _group_exists,
#     create as _group_create,
# )
# from burlap.files import uncommented_lines
# from burlap.utils import run_as_root
from burlap.decorators import task

 
_SALT_CHARS = string.ascii_letters + string.digits + './'

 
# # def create(name, comment=None, home=None, create_home=None, skeleton_dir=None,
# #            group=None, create_group=True, extra_groups=None, password=None,
# #            system=False, shell=None, uid=None, ssh_public_keys=None,
# #            non_unique=False):
# #     """
# #     Create a new user and its home directory.
# # 
# #     If *create_home* is ``None`` (the default), a home directory will be
# #     created for normal users, but not for system users.
# #     You can override the default behaviour by setting *create_home* to
# #     ``True`` or ``False``.
# # 
# #     If *system* is ``True``, the user will be a system account. Its UID
# #     will be chosen in a specific range, and it will not have a home
# #     directory, unless you explicitely set *create_home* to ``True``.
# # 
# #     If *shell* is ``None``, the user's login shell will be the system's
# #     default login shell (usually ``/bin/sh``).
# # 
# #     *ssh_public_keys* can be a (local) filename or a list of (local)
# #     filenames of public keys that should be added to the user's SSH
# #     authorized keys (see :py:func:`burlap.user.add_ssh_public_keys`).
# # 
# #     Example::
# # 
# #         import burlap
# # 
# #         if not burlap.user.exists('alice'):
# #             burlap.user.create('alice')
# # 
# #         with cd('/home/alice'):
# #             # ...
# # 
# #     """
# # 
# #     # Note that we use useradd (and not adduser), as it is the most
# #     # portable command to create users across various distributions:
# #     # http://refspecs.linuxbase.org/LSB_4.1.0/LSB-Core-generic/LSB-Core-generic/useradd.html
# # 
# #     args = []
# #     if comment:
# #         args.append('-c %s' % quote(comment))
# #     if home:
# #         args.append('-d %s' % quote(home))
# #     if group:
# #         args.append('-g %s' % quote(group))
# #         if create_group:
# #             if not _group_exists(group):
# #                 _group_create(group)
# #     if extra_groups:
# #         groups = ','.join(quote(group) for group in extra_groups)
# #         args.append('-G %s' % groups)
# # 
# #     if create_home is None:
# #         create_home = not system
# #     if create_home is True:
# #         args.append('-m')
# #     elif create_home is False:
# #         args.append('-M')
# # 
# #     if skeleton_dir:
# #         args.append('-k %s' % quote(skeleton_dir))
# #     if password:
# #         crypted_password = _crypt_password(password)
# #         args.append('-p %s' % quote(crypted_password))
# #     if system:
# #         args.append('-r')
# #     if shell:
# #         args.append('-s %s' % quote(shell))
# #     if uid:
# #         args.append('-u %s' % uid)
# #         if non_unique:
# #             args.append('-o')
# #     args.append(name)
# #     args = ' '.join(args)
# #     run_as_root('useradd %s' % args)
# # 
# #     if ssh_public_keys:
# #         if isinstance(ssh_public_keys, basestring):
# #             ssh_public_keys = [ssh_public_keys]
# #         add_ssh_public_keys(name, ssh_public_keys)
# 
# 
# def modify(name, comment=None, home=None, move_current_home=False, group=None,
#            extra_groups=None, login_name=None, password=None, shell=None,
#            uid=None, ssh_public_keys=None, non_unique=False):
#     """
#     Modify an existing user.
# 
#     *ssh_public_keys* can be a (local) filename or a list of (local)
#     filenames of public keys that should be added to the user's SSH
#     authorized keys (see :py:func:`burlap.user.add_ssh_public_keys`).
# 
#     Example::
# 
#         import burlap
# 
#         if burlap.user.exists('alice'):
#             burlap.user.modify('alice', shell='/bin/sh')
# 
#     """
# 
#     args = []
#     if comment:
#         args.append('-c %s' % quote(comment))
#     if home:
#         args.append('-d %s' % quote(home))
#         if move_current_home:
#             args.append('-m')
#     if group:
#         args.append('-g %s' % quote(group))
#     if extra_groups:
#         groups = ','.join(quote(group) for group in extra_groups)
#         args.append('-G %s' % groups)
#     if login_name:
#         args.append('-l %s' % quote(login_name))
#     if password:
#         crypted_password = _crypt_password(password)
#         args.append('-p %s' % quote(crypted_password))
#     if shell:
#         args.append('-s %s' % quote(shell))
#     if uid:
#         args.append('-u %s' % quote(uid))
#         if non_unique:
#             args.append('-o')
# 
#     if args:
#         args.append(name)
#         args = ' '.join(args)
#         run_as_root('usermod %s' % args)
# 
#     if ssh_public_keys:
#         if isinstance(ssh_public_keys, basestring):
#             ssh_public_keys = [ssh_public_keys]
#         add_ssh_public_keys(name, ssh_public_keys)
# 
# 
# def home_directory(name):
#     """
#     Get the absolute path to the user's home directory
# 
#     Example::
# 
#         import burlap
# 
#         home = burlap.user.home_directory('alice')
# 
#     """
#     with settings(hide('running', 'stdout')):
#         return run('echo ~' + name)
# 
# 
# def local_home_directory(name=''):
#     """
#     Get the absolute path to the local user's home directory
# 
#     Example::
# 
#         import burlap
# 
#         local_home = burlap.user.local_home_directory()
# 
#     """
#     with settings(hide('running', 'stdout')):
#         return local('echo ~' + name, capture=True)
# 
# 
# def authorized_keys(name):
#     """
#     Get the list of authorized SSH public keys for the user
#     """
# 
#     ssh_dir = posixpath.join(home_directory(name), '.ssh')
#     authorized_keys_filename = posixpath.join(ssh_dir, 'authorized_keys')
# 
#     return uncommented_lines(authorized_keys_filename, use_sudo=True)
# 
# 
# def add_ssh_public_key(name, filename):
#     """
#     Add a public key to the user's authorized SSH keys.
# 
#     *filename* must be the local filename of a public key that should be
#     added to the user's SSH authorized keys.
# 
#     Example::
# 
#         import burlap
# 
#         burlap.user.add_ssh_public_key('alice', '~/.ssh/id_rsa.pub')
# 
#     """
# 
#     add_ssh_public_keys(name, [filename])
# 
# 
# def add_ssh_public_keys(name, filenames):
#     """
#     Add multiple public keys to the user's authorized SSH keys.
# 
#     *filenames* must be a list of local filenames of public keys that
#     should be added to the user's SSH authorized keys.
# 
#     Example::
# 
#         import burlap
# 
#         burlap.user.add_ssh_public_keys('alice', [
#             '~/.ssh/id1_rsa.pub',
#             '~/.ssh/id2_rsa.pub',
#         ])
# 
#     """
# 
#     from burlap.require.files import (
#         directory as _require_directory,
#         file as _require_file,
#     )
# 
#     ssh_dir = posixpath.join(home_directory(name), '.ssh')
#     _require_directory(ssh_dir, mode='700', owner=name, use_sudo=True)
# 
#     authorized_keys_filename = posixpath.join(ssh_dir, 'authorized_keys')
#     _require_file(authorized_keys_filename, mode='600', owner=name,
#                   use_sudo=True)
# 
#     for filename in filenames:
# 
#         with open(filename) as public_key_file:
#             public_key = public_key_file.read().strip()
# 
#         # we don't use fabric.contrib.files.append() as it's buggy
#         if public_key not in authorized_keys(name):
#             sudo('echo %s >>%s' % (quote(public_key),
#                                    quote(authorized_keys_filename)))
# 
# 
# def add_host_keys(name, hostname):
#     """
#     Add all public keys of a host to the user's SSH known hosts file
#     """
# 
#     from burlap.require.files import (
#         directory as _require_directory,
#         file as _require_file,
#     )
# 
#     ssh_dir = posixpath.join(home_directory(name), '.ssh')
#     _require_directory(ssh_dir, mode='700', owner=name, use_sudo=True)
# 
#     known_hosts_filename = posixpath.join(ssh_dir, 'known_hosts')
#     _require_file(known_hosts_filename, mode='644', owner=name, use_sudo=True)
# 
#     known_hosts = uncommented_lines(known_hosts_filename, use_sudo=True)
# 
#     with hide('running', 'stdout'):
#         res = run('ssh-keyscan -t rsa,dsa %s 2>/dev/null' % hostname)
#     for host_key in res.splitlines():
#         if host_key not in known_hosts:
#             sudo('echo %s >>%s' % (quote(host_key),
#                                    quote(known_hosts_filename)))


#DEPRECATED
# def compare_manifest(old):
#     old = old or {}
#     methods = []
#     pre = []
#     
#     # Handle SSH key specification change.
#     old_key = (old.get('user_key_type'), old.get('user_key_bits'), old.get('user_key_filename'))
#     new_key = (env.get('user_key_type'), env.get('user_key_bits'), env.get('user_key_filename'))
#     if old_key != new_key:
#         methods.append(QueuedCommand('user.generate_keys', pre=pre))
#     
#     # Handle username change.
#     force_togroups = False
#     force_passwordless = env.user_passwordless and old.get('user_passwordless') != env.user_passwordless
#     if old.get('user') != env.user:
#         force_togroups = True
#         force_passwordless = env.user_passwordless
#         methods.append(QueuedCommand('user.create', kwargs=dict(username=env.user), pre=pre))
#     
#     # Handle user group change.
#     if force_togroups or old.get('user_groups') != env.user_groups:
#         methods.append(QueuedCommand('user.togroups', kwargs=dict(user=env.user, groups=env.user_groups), pre=pre))
#     
#     # Handle switch to passwordless access.
#     #TODO:support different username used for creating passworless user?
#     if force_passwordless:
#         methods.append(QueuedCommand('user.passwordless', kwargs=dict(username=env.user, pubkey=env.key_filename), pre=pre))
#         
#     #TODO: Handle switch from passwordless access? Remove old SSH key from remote and local caches?
#     
#     return methods

CAT_KEY = 'cat-key'
UPLOAD_KEY = 'upload-key'

def _crypt_password(password):
    from crypt import crypt
    random.seed()
    salt = ''
    for _ in range(2):
        salt += random.choice(_SALT_CHARS)
    crypted_password = crypt(password, salt)
    return crypted_password


class UserSatchel(Satchel):
    
    name = 'user'
    
    def set_defaults(self):
        
        self.env.tmp_sudoers_fn = '/tmp/sudoers'
        
        self.env.key_type = 'rsa' # e.g. rsa|dsa
        self.env.key_bits = 2048 # e.g. 1024, 2048, or 4096
        self.env.key_filename_template = 'roles/{ROLE}/{host}-{username}.pem'
        self.env.key_perms = '600'
        
        self.env.home_template = '/home/{username}'
        #self.env.passwordless = True
        self.env.groups = {} # {username:[groups]}
        self.env.passwordless = {} # {username:True/False}
        self.env.passwords = {} # {username: password}
        self.env.reset_passwords_on_first_login = {} # {username: true/false}
        
        self.env.default_passwords = {} # {username:password}
        
        self.env.passwordless_method = CAT_KEY

    @task
    def enter_password_change(self, username=None, old_password=None):
        """
        Responds to a forced password change via `passwd` prompts due to password expiration.
        """
        from fabric.state import connections
        from fabric.network import disconnect_all
        r = self.local_renderer
#         print('self.genv.user:', self.genv.user)
#         print('self.env.passwords:', self.env.passwords)
        r.genv.user = r.genv.user or username
        r.pc('Changing password for user {user} via interactive prompts.')
        r.env.old_password = r.env.default_passwords[self.genv.user]
#         print('self.genv.user:', self.genv.user)
#         print('self.env.passwords:', self.env.passwords)
        r.env.new_password = self.env.passwords[self.genv.user]
        if old_password:
            r.env.old_password = old_password 
        prompts = {
            '(current) UNIX password: ': r.env.old_password,
            'Enter new UNIX password: ': r.env.new_password,
            'Retype new UNIX password: ': r.env.new_password,
            #"Login password for '%s': " % r.genv.user: r.env.new_password,
#             "Login password for '%s': " % r.genv.user: r.env.old_password,
        }
        print('prompts:', prompts)
        
        r.env.password = r.env.old_password
        with self.settings(warn_only=True):
            ret = r._local("sshpass -p '{password}' ssh -o StrictHostKeyChecking=no {user}@{host_string} echo hello", capture=True)
            #code 1 = good password, but prompts needed
            #code 5 = bad password
            #code 6 = good password, but host public key is unknown
        if ret.return_code in (1, 6) or 'hello' in ret:
            # Login succeeded, so we haven't yet changed the password, so use the default password.
            self.genv.password = r.env.old_password
        elif self.genv.user in self.genv.user_passwords:
            # Otherwise, use the password or key set in the config.
            self.genv.password = r.env.new_password
        else:
            # Default password fails and there's no current password, so clear.
            self.genv.password = None
        print('using password:', self.genv.password)
        
        # Note, the correct current password should be set in host.initrole(), not here.
        #r.genv.password = r.env.new_password
        #r.genv.password = r.env.new_password
        with self.settings(prompts=prompts):
            ret = r._run('echo checking for expired password')
            print('ret:[%s]' % ret)
            do_disconnect = 'passwd: password updated successfully' in ret
            print('do_disconnect:', do_disconnect)
            if do_disconnect:
                # We need to disconnect to reset the session or else Linux will again prompt
                # us to change our password.
                disconnect_all()
                
                # Further logins should require the new password.
                self.genv.password = r.env.new_password
                
    @task
    def configure_keyless(self):
        self.generate_keys()
        self.passwordless()

    @task
    def togroups(self, user, groups):
        """
        Adds the user to the given list of groups.
        """
        
        r = self.local_renderer
        
        if isinstance(groups, basestring):
            groups = [_.strip() for _ in groups.split(',') if _.strip()]
        for group in groups:
            r.env.username = user
            r.env.group = group
            r.sudo('groupadd --force {group}')
            r.sudo('adduser {username} {group}')

    @task
    def passwordless(self, username, pubkey):
        """
        Configures the user to use an SSL key without a password.
        Assumes you've run generate_keys() first.
        """
        
        r = self.local_renderer
        
        r.env.username = username
        r.env.pubkey = pubkey
        if not self.dryrun:
            assert os.path.isfile(r.env.pubkey), \
                'Public key file "%s" does not exist.' % (str(r.env.pubkey),)
        
        first = os.path.splitext(r.env.pubkey)[0]
        r.env.pubkey = first+'.pub'
        r.env.pemkey = first+'.pem'
        r.env.home = r.env.home_template.format(username=username)
        
        # Upload the SSH key.
        r.sudo('mkdir -p {home}/.ssh')
        r.sudo('chown -R {user}:{user} {home}/.ssh')
        
        if r.env.passwordless_method == UPLOAD_KEY:
            put_remote_paths = self.put(local_path=r.env.pubkey)
            r.env.put_remote_path = put_remote_paths[0]
            r.sudo('cat {put_remote_path} >> {home}/.ssh/authorized_keys')
            r.sudo('rm -f {put_remote_path}')
        elif r.env.passwordless_method == CAT_KEY:
            r.env.password = r.env.default_passwords.get(r.env.username, r.genv.password)
            if r.env.password:
                r.local("cat {pubkey} | sshpass -p '{password}' ssh {user}@{host_string} 'cat >> {home}/.ssh/authorized_keys'")
            else:
                r.local("cat {pubkey} | ssh {user}@{host_string} 'cat >> {home}/.ssh/authorized_keys'")
        else:
            raise NotImplementedError
        
        # Disable password.
        r.sudo('cp /etc/sudoers {tmp_sudoers_fn}')
        r.sudo('echo "{username} ALL=(ALL) NOPASSWD: ALL" >> {tmp_sudoers_fn}')
        r.sudo('sudo EDITOR="cp {tmp_sudoers_fn}" visudo')
        
        r.sudo('service ssh reload')
        
        print('You should now be able to login with:')
        r.env.host_string = self.genv.host_string or (self.genv.hosts and self.genv.hosts[0])#self.genv.hostname_hostname
        r.comment('\tssh -i {pemkey} {username}@{host_string}')

    @task
    def generate_keys(self, username, hostname):
        """
        Generates *.pem and *.pub key files suitable for setting up passwordless SSH.
        """
        
        r = self.local_renderer
        
        #r.env.key_filename = r.env.key_filename or env.key_filename
        #assert r.env.key_filename, 'r.env.key_filename or env.key_filename must be set. e.g. roles/role/app_name-role.pem'
        r.env.key_filename = self.env.key_filename_template.format(
            ROLE=self.genv.ROLE,
            host=hostname,
            username=username,
        )
        if os.path.isfile(r.env.key_filename):
            r.pc('Key file {key_filename} already exists. Skipping generation.'.format(**r.env))
        else:
            r.local("ssh-keygen -t {key_type} -b {key_bits} -f {key_filename} -N ''")
            r.local('chmod {key_perms} {key_filename}')
            if r.env.key_filename.endswith('.pem'):
                src = r.env.key_filename+'.pub'
                dst = (r.env.key_filename+'.pub').replace('.pem', '')
#                 print('generate_keys:', src, dst)
                r.env.src = src
                r.env.dst = dst
                r.local('mv {src} {dst}')
        return r.env.key_filename

    @task
    def exists(self, name):
        """
        Check if a user exists.
        """
        with self.settings(hide('running', 'stdout', 'warnings'), warn_only=True):
            return self.run('getent passwd %s' % name).succeeded

    @task
    def create(self, username, groups=None, uid=None, create_home=None, system=False, password=None):
        """
        Creates a user with the given username.
        """
        r = self.local_renderer
        r.env.username = username
        
        args = []
        
        if uid:
            args.append('-u %s' % uid)
            
        if create_home is None:
            create_home = not system
            
        if create_home is True:
            #args.append('-m')
            pass
        elif create_home is False:
            args.append('--no-create-home')
        
        if password is None:
            pass
        elif password:
            crypted_password = _crypt_password(password)
            args.append('-p %s' % quote(crypted_password))
        else:
            args.append('--disabled-password')
        
        args.append('--gecos ""')
        
        if system:
            args.append('--system')
            
        r.env.args = ' '.join(args)
        r.env.groups = (groups or '').strip()
        r.sudo('adduser {username} {groups} {args} || true')
    
    @task
    def expire_password(self, username):
        """
        Forces the user to change their password the next time they login.
        """
        r = self.local_renderer
        r.env.username = username
        r.sudo('chage -d 0 {username}')
    
    @task
    def configure(self):
        r = self.local_renderer
        
        lm = self.last_manifest
        lm_reset_passwords_on_first_login = lm.reset_passwords_on_first_login or {}
        lm_passwordless = lm.passwordless or {}
        lm_passwords = lm.passwords or {}
        
        # Make one-time password changes.
        just_changed = set()
        for username, ret in r.env.reset_passwords_on_first_login.items():
            if ret and not lm_reset_passwords_on_first_login.get(username):
                self.enter_password_change(username)
                just_changed.add(username)
        
        # Make passwordless logins.
        for username, is_passwordless in r.env.passwordless.items():
            if is_passwordless:
                if not lm_passwordless.get(username):
                    # If this user is passwordless, and we've not already created a passwordless
                    # login for them, then create one.
                    pubkey = self.generate_keys(username=username, host=self.genv.hostname_hostname)
                    self.passwordless(username=username, pubkey=pubkey)
            else:
                #TODO:expire old SSH key?
                pass
        
        # Update passwords.
        for username, password in r.env.passwords.items():
            if username in just_changed:
                continue
            if lm_passwords.get(username) != password:
                r.env.username = username
                r.env.password = password
                r.sudo('echo "{username}:{password}"|chpasswd')
        
        # Set groups.
        for username, groups in r.env.groups.items():
            self.togroups(username, groups)

user = UserSatchel()



from __future__ import print_function

import os
import re
import sys
import datetime
import tempfile
import json
import functools
import traceback
from collections import defaultdict
from pprint import pprint

import yaml

from fabric.api import (
    env, runs_once, sudo as _sudo, get as _get,
)
import fabric.contrib.files
import fabric.api

from burlap import common
from burlap.common import (
    local_or_dryrun,
    put_or_dryrun,
    sudo_or_dryrun,
)
from burlap.decorators import task_or_dryrun
from burlap import exceptions

STORAGE_LOCAL = 'local'
STORAGE_REMOTE = 'remote'
STORAGES = (
    STORAGE_LOCAL,
    STORAGE_REMOTE,
)

default_remote_path = '/var/local/burlap'

# Prevent globals from being reset by duplicate imports.
if not 'plan_init' in env:
    env.plan_init = True
    env.plan_root = None
    env.plan_originals = {}
    env.plan_storage = STORAGE_REMOTE
    env.plan_lockfile_path = '/var/lock/burlap_deploy.lock'
_originals = env.plan_originals

env.plan = None
env.plan_data_dir = '%(burlap_data_dir)s/plans'
env.plan_digits = 3

RUN = 'run'
SUDO = 'sudo'
LOCAL = 'local'
PUT = 'put'
PLAN_METHODS = [RUN, SUDO, LOCAL, PUT]

INITIAL = 'initial'

_fs_cache = defaultdict(dict) # {func_name:{path:ret}}

def make_dir(d):
    if d not in _fs_cache['make_dir']:
        if env.plan_storage == STORAGE_REMOTE:
            sudo_or_dryrun('mkdir -p "%s"' % d)
        else:
            if not os.path.isdir(d):
                os.makedirs(d)
        _fs_cache['make_dir'][d] = True
    return _fs_cache['make_dir'][d]

@task_or_dryrun
def list_dir(d):
    if d not in _fs_cache['list_dir']:
        verbose = common.get_verbose()
        if env.plan_storage == STORAGE_REMOTE:
            #output = sudo_or_dryrun('ls "%s"' % d)
            output = _sudo('ls "%s"' % d)
            output = output.split()
            if verbose:
                print('output:', output)
            ret = output
        else:
            ret = os.listdir(d)
        _fs_cache['list_dir'][d] = ret
    return _fs_cache['list_dir'][d]

@task_or_dryrun
def is_dir(d):
    if d not in _fs_cache['is_dir']:
        verbose = common.get_verbose()
        if env.plan_storage == STORAGE_REMOTE:
            cmd = 'if [ -d "%s" ]; then echo 1; else echo 0; fi' % d
            output = _sudo(cmd)
            if verbose:
                print('output:', output)
            #ret = int(output)
            ret = int(re.findall(r'^[0-9]+$', output, flags=re.DOTALL|re.I|re.M)[0])
        else:
            ret = os.path.isdir(d)
        _fs_cache['is_dir'][d] = ret
    return _fs_cache['is_dir'][d]

@task_or_dryrun
def is_file(fqfn):
    if fqfn not in _fs_cache['is_file']:
        verbose = common.get_verbose()
        if env.plan_storage == STORAGE_REMOTE:
            cmd = 'if [ -f "%s" ]; then echo 1; else echo 0; fi' % fqfn
            output = _sudo(cmd)
            if verbose:
                print('output:', output)
            ret = int(re.findall(r'^[0-9]+$', output, flags=re.DOTALL|re.I|re.M)[0])
        else:
            ret = os.path.isfile(fqfn)
        _fs_cache['is_file'][fqfn] = ret
    return _fs_cache['is_file'][fqfn]

# class Singleton(type):
#     def __init__(cls, name, bases, dict):
#         print('singleton.init')
#         super(Singleton, cls).__init__(name, bases, dict)
#         cls.instance = None
#  
#     def __call__(cls, *args, **kw):
#         print('singleton.call')
#         return super(Singleton, cls).__call__(*args, **kw)
#         if cls.instance is None:
#             cls.instance = super(Singleton, cls).__call__(*args, **kw)
#         return cls.instance

class RemoteFile(object):
    """
    A helper class for allowing a remote file to be read and written locally
    while still ultimately being saved remotely.
    """
    
#     __metaclass__ = Singleton
    
    _file_cache = {} # {fqfn, obj}
    
    #TODO:use meta-class instead?
    #http://stackoverflow.com/questions/31875/is-there-a-simple-elegant-way-to-define-singletons-in-python/33201#33201
    
    def __new__(cls, fqfn, *args, **kwargs):
        # Remember and cache every class instance per unique file name.
        if fqfn not in cls._file_cache:
#             print('creating new instance:', fqfn)
            cls._file_cache[fqfn] = super(RemoteFile, cls).__new__(cls, fqfn, *args, **kwargs)
#         else:
#             print('using cache:', fqfn)
        return cls._file_cache[fqfn]
    
    def __init__(self, fqfn, mode='r'):
        super(RemoteFile, self).__init__() # causes instantiation error?
        
        assert mode in ('r', 'w', 'a'), 'Invalid mode: %s' % mode
        
        self.mode = mode
            
        # Due to the singleton-nature of __new__, this may be called multiple times,
        # so we check for and avoid duplicate calls.
        if not hasattr(self, 'fqfn'):
            
            self.fqfn = fqfn
            self.content = ''
            self.fresh = True
            
            if mode in 'ra':
    
                _, tmp_fn = tempfile.mkstemp()
                os.remove(tmp_fn)
                ret = _get(remote_path=fqfn, local_path=tmp_fn, use_sudo=True)
                #ret = get_or_dryrun(remote_path=fqfn, local_path=tmp_fn, use_sudo=True)
#                 print('ret:', ret)
                _fn = ret[0]
#                 print('reading:', _fn)
                fin = open(_fn, 'rb')
#                 print('reading2:', _fn)
                self.content = fin.read()
#                 print('closing')
                fin.close()
                #print('removing:', tmp_fn)
                os.remove(tmp_fn)#TODO:memory leak?
#                 print('done init load')
                
            if mode in 'wa':
                
                # Update file system cache.
                _fs_cache['is_file'][fqfn] = True
                
#         print('done init all')

    def write(self, s):
        assert self.mode in 'wa', 'File must be in write-mode.'
        self.content += s
        self.fresh = False
        # Note, flush() must to be called to actually write this.

    def read(self, *args, **kwargs):
        return self.content
    
    def readlines(self):
        return self.content.split('\n')
    
    def flush(self):
        if self.fresh:
            return
        
        print('Flushing contents to remote file.')
        _, tmp_fn = tempfile.mkstemp()
        os.remove(tmp_fn)
        fout = open(tmp_fn, 'w')
        fout.write(self.content)
        fout.close()
        put_or_dryrun(
            local_path=tmp_fn,
            remote_path=self.fqfn,
            use_sudo=True)
        os.remove(tmp_fn)#TODO:memory leak?
        self.fresh = True
        
        # Update file system cache.
        _fs_cache['is_file'][self.fqfn] = True
        
    def close(self):
        print('Closing remote file.')
        self.flush()

def open_file(fqfn, mode='r'):
    verbose = common.get_verbose()
    if env.plan_storage == STORAGE_REMOTE:
        return RemoteFile(fqfn, mode)
    else:
        return open(fqfn, mode)

def init_plan_data_dir():
    common.init_burlap_data_dir()
    d = env.plan_data_dir % env
    make_dir(d)
    return d

class Colors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

def fail(s):
    return Colors.FAIL + str(s) + Colors.ENDC

def success(s):
    return Colors.OKGREEN + str(s) + Colors.ENDC

def ongoing(s):
    return Colors.WARNING + str(s) + Colors.ENDC

def iter_plan_names(role=None):
    d = get_plan_dir(role=role)
    try:
        assert is_dir(d), 'Plan directory %s does not exist.' % d
    except AssertionError:
        if common.get_dryrun():
            # During dryrun, and the directory is missing, assume the host has been reset
            # and there are no prior plan files.
            return
        else:
            raise
    for name in sorted(list_dir(d)):
        fqfn = os.path.join(d, name)
        if not is_dir(fqfn):
            continue
        yield name

def get_thumbprint_path(role, name):
    d = get_plan_dir(role=role, name=name)
    d = os.path.join(d, 'thumbprints')
    make_dir(d)
    return d

def get_thumbprint_filename():
    return 'thumbprint'
        
class Step(object):
    """
    A single piece of a plan.
    """
    
    def __init__(self, command, host, method, user=None, key=None, args=None, kwargs=None):
        
        args = args or []
        kargs = kwargs or {}
        
        self.command = command
        self.host = host
        self.user = user
        self.method = method
        
        assert method in PLAN_METHODS
        
        self.key = key
        # The value entity attributes are set to as a result
        # of executing this step.
        self.args = args
        self.kwargs = kwargs
    
    @classmethod
    def from_line(cls, line):
        matches = re.findall(
            r'\[(?P<user>[^@]+)@(?P<host>[^\]]+)(?P<extra>{[^}]+})?]\s+(?P<method>[^\:]+):\s+(?P<command>.*?)$',
            line, flags=re.I|re.DOTALL)
        assert matches
        #print(matches)
        kwargs = dict(zip(['user', 'host', 'extra', 'method', 'command'], matches[0]))
        
        extra = {}
        if kwargs['extra']:
            extra = json.loads(kwargs['extra'])
        del kwargs['extra']
        
        if 'key' in extra:
            key = extra['key']
            del extra['key']
            kwargs['key'] = key
           
        step = Step(**kwargs)
        return step
    
    def execute(self):
        method = getattr(common, '%s_or_dryrun' % self.method)
        env.user = self.user
        env.host_string = self.host
        if self.key:
            env.key_filename = self.key
        else:
            env.key_filename = None
        method(self.command)
    
    def __str__(self):
        user_str = '%s@' % (self.user) if self.user else ''
        return '[%s%s] %s %s' % (
            user_str,
            self.host,
            self.method+':',
            self.command,
        )
    
    def __repr__(self):
        return str(self.__dict__)

HISTORY_HEADERS = ['step', 'start', 'end']

def get_plan_dir(role, name=None):
    if name:
        d = os.path.join(init_plan_data_dir(), role or env.ROLE, name)
    else:
        d = os.path.join(init_plan_data_dir(), role or env.ROLE)
    make_dir(d)
    return d

class Plan(object):
    """
    A sequence of steps for accomplishing a state change.
    """
    
    def __init__(self, name, role=None):
        
        self.verbose = verbose = common.get_verbose()
        
        self.name = name
        
        self.role = role or env.ROLE
        
        self.vprint('init plan dir')
        self.plan_dir = get_plan_dir(role, name)
        make_dir(self.plan_dir)
        assert is_dir(self.plan_dir)
        
        self.vprint('init plan history dir')
        self.plan_dir_history = os.path.join(self.plan_dir, 'history')
        if not is_file(self.plan_dir_history):
            fout = open_file(self.plan_dir_history, 'w')
            fout.write(','.join(HISTORY_HEADERS))
            fout.close()
        self.vprint('loading plan history')
        
        self.load_history()
        
        self.vprint('init plan index')
        self.plan_dir_index = os.path.join(self.plan_dir, 'index')
        if not is_file(self.plan_dir_index):
            fout = open_file(self.plan_dir_index, 'w')
            fout.write(str(0))
            fout.close()
        self.vprint('loading plan index')
        self.load_index()
        
        self.vprint('init plan steps')
        self.plan_dir_steps = os.path.join(self.plan_dir, 'steps')
        if not is_file(self.plan_dir_steps):
            fout = open_file(self.plan_dir_steps, 'w')
            fout.write('')
            fout.close()
        self.vprint('loading plan steps')
        self.load_steps()
        
        self.vprint('init plan hosts')
        self.plan_dir_hosts = os.path.join(self.plan_dir, 'hosts')
        if self.role == env.ROLE and not is_file(self.plan_dir_hosts):
            fout = open_file(self.plan_dir_hosts, 'w')
            fout.write('\n'.join(sorted(env.hosts)))
            fout.close()
        self.vprint('loading plan hosts')
        self.load_hosts()
        
        #self.plan_thumbprint_fn = os.path.join(self.plan_dir, 'thumbprint')
        
        self.vprint('plan init done')

    def vprint(self, *args, **kwargs):
        """
        When verbose is set, acts like the normal print() function.
        Otherwise, does nothing.
        """
        if common.get_verbose():
            print(*args, **kwargs)

    def __cmp__(self, other):
        if not isinstance(other, Plan):
            return NotImplemented
        return cmp((self.name, self.role), (other.name, other.role))
    
    def __unicode__(self):
        return unicode(self.name)
    
    def __repr__(self):
        return u'<%s: %s>' % (type(self).__name__, unicode(self))
    
    def is_complete(self):
        #return self.percent_complete == 100
        return self.all_hosts_thumbprinted and self.index == len(self._steps)
    
    @property
    def number(self):
        try:
            return int(re.findall('^[0-9]+', self.name)[0])
        except IndexError:
            #print('No number in "%s"' % self.name
            return 0
    
    def failed(self):
        return False #TODO
    
    def load_hosts(self):
        self.hosts = []
        if self.verbose: print('loading hosts, opening')
        fin = open_file(self.plan_dir_hosts, 'r')
        if self.verbose: print('loading hosts, readlines')
        lines = fin.readlines()
        if self.verbose: print('loading hosts, lines:', lines)
        for line in lines:
            if not line.strip():
                continue
            self.hosts.append(line.strip())
        if self.verbose: print('loading hosts, done')
    
    @property
    def all_hosts_thumbprinted(self):
        for host in self.hosts:
            fn = self.get_thumbprint_filename(host)
            if not is_file(fn):
                return False
        return True
    
    def load_history(self):
        pass
    
    def get_thumbprint_filename(self, host_string):
        d = os.path.join(self.plan_dir, 'thumbprints')
        make_dir(d)
        fn = os.path.join(d, env.host_string)
        return fn
    
    @property
    def thumbprint(self):
        verbose = common.get_verbose()
        self.vprint('plan.thumbprint')
        fn = self.get_thumbprint_filename(env.host_string)
        self.vprint('plan.thumbprint.fn:', fn)
        content = open_file(fn).read()
        self.vprint('plan.thumbprint.yaml.raw:', content)
        data = yaml.load(content)
        self.vprint('plan.thumbprint.yaml.data:', data)
        return data
    
    @thumbprint.setter
    def thumbprint(self, data):
        assert isinstance(data, dict)
        if not common.get_dryrun():
            fout = open_file(self.get_thumbprint_filename(env.host_string), 'w')
            yaml.dump(data, fout, default_flow_style=False, indent=4)
            fout.flush()
    
    def record_thumbprint(self, only_components=None):
        """
        Creates a thumbprint file for the current host in the current role and name.
        """
        only_components = only_components or []
        data = get_current_thumbprint(role=self.role, name=self.name, only_components=only_components)
        print('Recording thumbprint for host %s with deployment %s on %s.' \
            % (env.host_string, self.name, self.role))
        self.thumbprint = data
    
    @property
    def remaining_step_count(self):
        return len(self._steps) - self.index
    
    def add_history(self, index, start, end):
        fout = open_file(self.plan_dir_history, 'a')
        fout.write('%s,%s,%s\n' % (index, start, end))
        fout.flush()
        fout.close()
    
    def load_index(self):
        self._index = int(open_file(self.plan_dir_index).read().strip())
    
    @property
    def index(self):
        return self._index
    
    def is_initial(self):
        return set(self.name) == set(['0'])
    
    @property
    def percent_complete(self):
        if self.is_initial():
            return 100
        if not self._steps:
            return 100
        return self.index/float(len(self._steps))*100
    
    @index.setter
    def index(self, v):
        self._index = int(v)
        fout = open_file(self.plan_dir_index, 'w')
        fout.write(str(self._index))
        fout.flush()
        fout.close()
    
    def load_steps(self):
        self._steps = []
        lines = open_file(self.plan_dir_steps).readlines()
        for line in lines:
            line = line.strip()
            if not line:
                continue
            elif '] executing task ' in line.lower():
                continue
            elif line.lower().startswith('done'):
                continue
            s = Step.from_line(line)
            self.add_step(s)

    @classmethod
    def get_or_create_next(cls, role=None, last_plan=None):
        role = role or env.ROLE
        last_plan = last_plan or get_last_plan(role=role)
        if last_plan:
            number = last_plan.number + 1
        else:
            number = 0
        assert len(str(number)) <= env.plan_digits, \
            'Too many deployments. Truncate existing or increase `plan_digits`.'
        plan = Plan(role=role, name=('%0'+str(env.plan_digits)+'i') % number)
        return plan

    @classmethod
    def load(cls, name, role=None):
        verbose = common.get_verbose()
        if verbose:
            print('loading plan:', name)
        plan = cls(name, role=role)
        return plan
    
    def add_step(self, s):
        assert isinstance(s, Step)
        self._steps.append(s)
    
    @property
    def steps(self):
        return list(self._steps)
    
    def clear(self):
        self.index = 0
        self._steps = []
    
    def execute(self, i=None, j=None):
        i = i or self.index
        steps_ran = []
        for step_i, step in enumerate(self.steps):
            if step_i < i:
                continue
                
            # Run command.
            t0 = datetime.datetime.utcnow().isoformat()
            step.execute()
            t1 = datetime.datetime.utcnow().isoformat()
            steps_ran.append(step)
            
            # Record success.
            if not common.get_dryrun():
                self.index = step_i + 1
                self.add_history(self.index, t0, t1)
            
            if j is not None and step_i >= j:
                break
                
        return steps_ran

#DEPRECATED
@task_or_dryrun
def record(name):
    common.set_dryrun(1)
    env.plan = Plan(name=name)

#DEPRECATED
@task_or_dryrun
def execute(name):
    verbose = common.get_verbose()
    plan = Plan.load(name, verbose=int(verbose))
    if verbose:
        if plan.is_complete():
            print('Execution of plan %s is complete.' % (plan.name,), file=sys.stderr)
        else:
            print('Execution of plan %s is %.02f%% complete.' % (plan.name, plan.percent_complete), file=sys.stderr)
    steps = []
    if not plan.is_complete():
        steps = plan.execute()
        if verbose:
            if plan.is_complete():
                print('Execution of plan %s is complete.' % (plan.name,), file=sys.stderr)
            else:
                print('Execution of plan %s is %.02f%% complete.' % (plan.name, plan.percent_complete), file=sys.stderr)
    if verbose:
        print('Executed %i steps.' % (len(steps),), file=sys.stderr)

def get_last_completed_plan():
    """
    Returns the last plan completed.
    """
    verbose = common.get_verbose()
    if verbose:
        print('get_last_completed_plan')
    for _name in reversed(sorted(list(iter_plan_names()))):
        plan = Plan.load(_name)
        if verbose:
            print('plan:', plan.name)
            print('plan.completed:', plan.is_complete())
        if plan.is_complete():
            return plan
            
def get_last_plan(role=None):
    """
    Returns the last plan created.
    """
    for _name in reversed(sorted(list(iter_plan_names(role=role)))):
        plan = Plan.load(_name, role=role)
        return plan

@task_or_dryrun
def has_outstanding_plans():
    """
    Returns true if there are plans for this role that have not been executed.
    """
    verbose = common.get_verbose()
    last_completed = get_last_completed_plan()
    if verbose:
        print('last_completed plan:', last_completed)
    last = get_last_plan()
    if verbose:
        print('last plan:', last)
        print('eq:', last == last_completed)
    return last != last_completed

@task_or_dryrun
@runs_once
def status(name=None):
    """
    Reports the status of any pending plans for the current role.
    """
    print('plan,complete,percent')
    for _name in iter_plan_names():
        #print(_name)
        plan = Plan.load(_name)
        output = '%s,%s,%s' % (_name, int(plan.is_complete()), plan.percent_complete)
        if plan.is_complete():
            output = success(output)
        elif plan.failed:
            output = fail(output)
        else:
            output = ongoing(output)
        print(output)

def get_current_thumbprint(role=None, name=None, reraise=0, only_components=None):
    """
    Retrieves a snapshot of the current code state.
    """
    if name == INITIAL:
        name = '0'*env.plan_digits
    
    last = get_last_thumbprint()
    only_components = only_components or []
    only_components = [_.upper() for _ in only_components]
    data = {} # {component:data}
    manifest_data = (last and last.copy()) or {}
#     print('manifest_data:', manifest_data.keys())
#     print('only_components:', only_components)
#     raw_input('enter')
    for component_name, func in sorted(common.manifest_recorder.iteritems()):
        component_name = component_name.upper()
        #print('component_name:', component_name)
        
        if only_components and component_name not in only_components:
            if common.get_verbose():
                print('Skipping ignored component:', component_name)
            continue
            
        if component_name.lower() not in env.services:
            if common.get_verbose():
                print('Skipping unused component:', component_name)
            continue
            
        try:
            manifest_data[component_name] = func()
#             print('manifest:', component_name, manifest_data[component_name])
        except exceptions.AbortDeployment as e:
            raise
        except Exception as e:
            if int(reraise):
                raise
            print(traceback.format_exc(), file=sys.stderr)
        
    return manifest_data

@task_or_dryrun
def get_last_thumbprint():
    """
    Returns thumbprint from the last complete deployment.
    """
    verbose = common.get_verbose()
    plan = get_last_completed_plan()
    if verbose and plan: print('get_last_thumbprint.last completed plan:', plan.name)
    last_thumbprint = (plan and plan.thumbprint) or {}
    if verbose: print('get_last_thumbprint.last_thumbprint:', last_thumbprint)
    return last_thumbprint

def iter_thumbprint_differences(only_components=None, local_verbose=0):
    only_components = only_components or []
    local_verbose = int(local_verbose)
    verbose = common.get_verbose() or local_verbose
    #if verbose: print('getting last thumbprint')
    last = get_last_thumbprint()
    #if verbose: print('getting current thumbprint')
    current = get_current_thumbprint()
    #if verbose: print('comparing thumbprints')
    for k in current:
        if only_components and k not in only_components:
#             print('iter_thumbprint_differences.skipping:', k)
            continue
#         print('iter:',k); raw_input('enter')
#         if verbose: print('iter_thumbprint_differences.NOT skipping:', k)
        if current[k] != last.get(k):
            if verbose:
                print('DIFFERENCE! k:', k, current[k], last.get(k))
                print('Current:')
                pprint(current[k], indent=4)
                print('Last:')
                pprint(last.get(k), indent=4)
            yield k, (last, current)
#     if verbose: print('iter_thumbprint_differences done')

@task_or_dryrun
def explain(name, **kwargs):
    #common.set_verbose(1)
    kwargs = kwargs or {}
    name = common.assert_valid_satchel(name)
    kwargs['only_components'] = [name]
    kwargs['local_verbose'] = 1
    diffs = dict(iter_thumbprint_differences(**kwargs))
    last, current = diffs.get(name, (None, None))
    if last is None and current is None:
        print('There are no differences.')
#     else:
#         last = last or {}
#         last.setdefault(name, {})
#         print('last:')
#         pprint(last[name], indent=4)
#         print('current:')
#         pprint(current[name], indent=4)

@task_or_dryrun
def show_diff(only=None):
    """
    Inspects differences between the last deployment and the current code state.
    """
    for k, (last, current) in iter_thumbprint_differences():
        if only and k.lower() != only.lower():
            continue
        print('Component %s has changed.' % k)
        last = last.get(k)
        current = current.get(k)
        if isinstance(last, dict) and isinstance(current, dict):
            for _k in set(last).union(current):
                _a = last.get(_k)
                _b = current.get(_k)
                if _a != _b:
                    print('DIFF: %s =' % _k, _a, _b)
        else:
            print('DIFF:', last, current)

@task_or_dryrun
def info():
    d = os.path.join(init_plan_data_dir(), env.ROLE)
    print('storage:', env.plan_storage)
    print('dir:', d)

@task_or_dryrun
def reset():
    """
    Deletes all recorded plan executions.
    This will cause the planner to think everything needs to be re-deployed.
    """
    d = os.path.join(init_plan_data_dir(), env.ROLE)
    if env.plan_storage == STORAGE_REMOTE:
        sudo_or_dryrun('rm -Rf "%s"' % d)
        sudo_or_dryrun('mkdir -p "%s"' % d)
    elif env.plan_storage == STORAGE_LOCAL:
        local_or_dryrun('rm -Rf "%s"' % d)
        local_or_dryrun('mkdir -p "%s"' % d)
    else:
        raise NotImplementedError
    
@task_or_dryrun
@runs_once
def truncate():
    """
    Compacts all deployment records into a single initial deployment.
    """
    reset()
    if not common.get_dryrun():
        fabric.api.execute(thumbprint, hosts=env.hosts)

@task_or_dryrun
def thumbprint(name=None, components=None):
    """
    Creates a manifest file for the current host, listing all current settings
    so that a future deployment can use it as a reference to perform an
    idempotent deployment.
    """
    
    only_components = components or []
    if isinstance(only_components, basestring):
        only_components = [_.strip() for _ in only_components.split(',') if _.strip()]
    
    if name:
        plan = Plan.load(name=name)
    else:
        plan = get_last_plan()
        print('last plan:', plan)
        if not plan:
            plan = Plan.get_or_create_next()
    plan.record_thumbprint(only_components=only_components)
    
@task_or_dryrun
@runs_once
def preview(**kwargs):
    """
    Lists the likely pending deployment steps.
    """
    return auto(preview=1, **kwargs)

def get_last_current_diffs(target_component):
    """
    Retrieves differing manifests between the current and last snapshot.
    """
    target_component = target_component.strip().upper()
    
    all_services = set(_.strip().upper() for _ in env.services)
    diffs = list(iter_thumbprint_differences())
    components = set()
    component_thumbprints = {}
    for component, (last, current) in diffs:
        if component not in all_services:
            continue
        component_thumbprints[component] = last, current
    
    print('component_thumbprints:', component_thumbprints.keys())
    last, current = component_thumbprints[target_component]
    return last, current

@task_or_dryrun
def auto(fake=0, preview=0, check_outstanding=1, components=None, explain=0):
    """
    Generates a plan based on the components that have changed since the last deployment.
    
    The overall steps ran for each host:
    
        1. create plan
        2. run plan
        3. create thumbprint
    
    fake := If true, generates the plan and records the run as successful, but does not apply any
        changes to the hosts.
    
    components := list of names of components found in the services list
    
    """
    
    explain = int(explain)
    only_components = components or []
    if isinstance(only_components, basestring):
        only_components = [_.strip().upper() for _ in only_components.split(',') if _.strip()]
    if only_components:
        print('Limiting deployment to components: %s' % only_components)
    
    def get_deploy_funcs(components):
        for component in components:
            
            if only_components and component not in only_components:
                continue
            
            funcs = common.manifest_deployers.get(component, [])
            for func_name in funcs:
                
                #TODO:remove this after burlap.* naming prefix bug fixed
                if func_name.startswith('burlap.'):
                    print('skipping %s' % func_name)
                    continue
                    
                takes_diff = common.manifest_deployers_takes_diff.get(func_name, False)
#                 print(func_name, takes_diff)
                
                if preview:
                    #print(success((' '*4)+func_name))
                    #continue
                    yield func_name, None
                else:
                    func = common.resolve_deployer(func_name)
                    last, current = component_thumbprints[component]
                    if not fake:
                        if takes_diff:
                            yield func_name, functools.partial(func, last=last, current=current)
                        else:
                            yield func_name, functools.partial(func)
    
    verbose = common.get_verbose()
    fake = int(fake)
    preview = int(preview)
    check_outstanding = int(check_outstanding)
    
    all_services = set(_.strip().upper() for _ in env.services)
    if verbose:
        print('&'*80)
        print('services:', env.services)
    
    last_plan = get_last_completed_plan()
    outstanding = has_outstanding_plans()
    if verbose:
        print('outstanding plans:', outstanding)
    if check_outstanding and outstanding:
        print(fail((
            'There are outstanding plans pending execution! '
            'Run `fab %s deploy.status` for details.\n'
            'To ignore these, re-run with :check_outstanding=0.'
        ) % env.ROLE))
        sys.exit(1)
    
    if verbose:
        print('iter_thumbprint_differences')
    diffs = list(iter_thumbprint_differences(only_components=only_components))
    if diffs:
        if verbose:
            print('Differences detected!')

    # Create plan.
    components = set()
    component_thumbprints = {}
    for component, (last, current) in diffs:
        if component not in all_services:
            print('ignoring component:', component)
            continue
#         if only_components and component not in only_components:
#             continue
        component_thumbprints[component] = last, current
        components.add(component)
    component_dependences = {}
    
    if verbose:
        print('all_services:', all_services)
        print('manifest_deployers_befores:', common.manifest_deployers_befores.keys())
        print('*'*80)
        print('all components:', components)
    
    all_components = set(common.all_satchels)
    if only_components and not all_components.issuperset(only_components):
        unknown_components = set(only_components).difference(all_components)
        raise Exception('Unknown components: %s' \
            % ', '.join(sorted(unknown_components)))
    
    for _c in components:
        if verbose:
            print('checking:', _c)
        deps = set(common.manifest_deployers_befores.get(_c, []))
        if verbose:
            print('deps0:', deps)
        deps = deps.intersection(components)
        if verbose:
            print('deps1:', deps)
        component_dependences[_c] = deps
        
    if verbose:
        print('dependencies:')
        for _c in component_dependences:
            print(_c, component_dependences[_c])
        
    components = list(common.topological_sort(component_dependences.items()))
#     print('components:',components)
#     raw_input('enter')
    plan_funcs = list(get_deploy_funcs(components))
    if components and plan_funcs:
        print('These components have changed:\n')
        for component in sorted(components):
            print((' '*4)+component)
        print('\nDeployment plan:\n')
        for func_name, _ in plan_funcs:
            print(success((' '*4)+func_name))
    else:
        print('Nothing to do!')
        return False
    
    # Execute plan. 
    if preview:
        print('\nTo execute this plan on all hosts run:\n\n    fab %s deploy.run' % env.ROLE)
        return components, plan_funcs
    else:
        with open('/tmp/burlap.progress', 'w') as fout:
            print('%s Beginning plan execution!' % (datetime.datetime.now(),), file=fout)
            fout.flush()
            for func_name, plan_func in plan_funcs:
                print('%s Executing step %s...' % (datetime.datetime.now(), func_name))
                print('%s Executing step %s...' % (datetime.datetime.now(), func_name), file=fout)
                fout.flush()
                if callable(plan_func):
                    plan_func()
                    
                    # Record this step complete.
                    if not only_components:
                        try:
                            thumbprint(components=func_name.split('.')[0])
                        except AssertionError:
                            # On new installs where the host is not yet present, this may fail.
                            pass
                        
                print('%s Done!' % (datetime.datetime.now(),), file=fout)
                fout.flush()
            print('%s Plan execution complete!' % (datetime.datetime.now(),), file=fout)
            fout.flush()
    
    # Create thumbprint.
    if not common.get_dryrun():
        plan = Plan.get_or_create_next(last_plan=last_plan)
        plan.record_thumbprint(only_components=only_components)

@task_or_dryrun
def run(*args, **kwargs):
    """
    Performs a full deployment.
    
    Parameters:
    
        components := name of satchel to limit deployment to
    """
    from burlap import notifier
    from burlap.common import all_satchels, get_satchel
    
    service = get_satchel('service')
    
    assume_yes = int(kwargs.pop('assume_yes', 0)) or int(kwargs.pop('yes', 0))
    fake = int(kwargs.get('fake', 0))
    
    # Allow satchels to configure connection parameters before we try contacting the hosts.
    #TODO:support ordering?
    for name, satchel in all_satchels.iteritems():
        if hasattr(satchel, 'deploy_pre_run'):
            satchel.deploy_pre_run()
    
    if env.host_string == env.hosts[0]:
        pending = preview(*args, **kwargs)
        if pending:
            # There are changes that need to be deployed, but confirm first with user.
            if not assume_yes \
            and not raw_input('\nBegin deployment? [yn] ').strip().lower().startswith('y'):
                sys.exit(1)
        else:
            # There are no changes pending, so abort all further tasks.
            sys.exit(1)
    
    if not fake:
        service.pre_deploy()
        
    kwargs['check_outstanding'] = 0
    auto(*args, **kwargs)
    
    if not fake:
        service.post_deploy()
        notifier.notify_post_deployment()

@task_or_dryrun
def test_remotefile():
    f = RemoteFile('/var/log/auth.log')
    f.read()
    print(id(f))
    print('-'*80)
    f = RemoteFile('/var/log/auth.log')
    print(id(f))
    


#!/usr/bin/env python
"""
Command line tool for initializing Burlap-structured Django projects.

e.g.

    burlap skel --name=myproject
"""
from __future__ import print_function

import argparse
import os
import sys
import re

_path = os.path.normpath(os.path.join(os.path.realpath(__file__), '../..'))
sys.path.insert(0, _path)

import burlap # pylint: disable=wrong-import-position
from burlap import common # pylint: disable=wrong-import-position

fabfile_template = os.path.join(
    os.path.dirname(burlap.__file__),
    'templates',
    'burlap',
    'fabfile.py.template',
)

ACTIONS = (
    SKEL,
    ADD_ROLE,
    CREATE_SATCHEL,
) = (
    'skel',
    'add-role',
    'create-satchel',
)

def md(d):
    if os.path.isdir(d):
        return
    os.makedirs(d)

def to_camelcase(value):
    value = re.sub(r'[^a-zA-Z0-9]+', ' ', value).strip()
    return ''.join(x.capitalize() for x in value.split(' '))

def init_django(args, virtualenv='.env'):
    
    site_name = args.project_name
    
    print('Initializing Django project...')
    if not os.path.isdir('src/%s' % site_name):
        print('Initializing base django project...')
        os.system('. %s/bin/activate; django-admin.py startproject %s src; deactivate' \
            % (virtualenv, site_name,))
        _settings_fn = 'src/%s_site/settings.py' % args.project_name
        _content = open(_settings_fn, 'r').read()
        _sites = '''SITE_{name_upper} = "{name_lower}"
    SITES = (
    SITE_{name_upper},
    )
    '''.format(
            name_upper=args.project_name.upper(),
            name_lower=args.project_name.lower(),
        )
        _top = []
        for _role in default_roles:
            _top.append("ROLE_%s = '%s'" % (_role.upper(), _role.lower()))
        _top.append('ROLES = (')
        for _role in default_roles:
            _top.append("    ROLE_%s," % (_role.upper(),))
        _top.append(')')
        _index = _content.find('"""\n\n')+4
        _bottom = '''
PROJECT_DIR = os.path.abspath(os.path.join(os.path.split(__file__)[0], '..', '..'))

STATIC_ROOT = os.path.join(PROJECT_DIR, 'static')

MEDIA_ROOT = os.path.join(PROJECT_DIR, 'media')
MEDIA_URL = '/media/'

STATICFILES_FINDERS = (
'django.contrib.staticfiles.finders.FileSystemFinder',
'django.contrib.staticfiles.finders.AppDirectoriesFinder',
)
TEMPLATE_LOADERS = (
'django.template.loaders.filesystem.Loader',
'django.template.loaders.app_directories.Loader',
)
TEMPLATE_DIRS = (
'%s/src/{app_name}/templates' % PROJECT_DIR
)
ADMIN_TITLE = '{app_name_title}'
ADMIN_TITLE_SIMPLE = '{app_name_simple}'
'''.format(
    app_name=args.project_name,
    app_name_title=args.project_name.title() + ' Administration',
    app_name_simple=args.project_name.title())
        open(_settings_fn, 'w').write(_content[:_index]+_sites+('\n'.join(_top))+_content[_index:]+_bottom)
    
    print('Creating Django helper scripts...')
    open('src/manage', 'w').write('''#!/bin/bash
# Helper script for ensuring we use the Python binary in our local
# virtual environment when calling management commands.
# Otherwise, we'd have to always run `. ../.env/bin/activate`, which can be
# annoying.
# Be sue to run `fab <role> pip.init` first in order to setup
# the target role's Python virtual environment.
DIR=`dirname $0`;
cd $DIR;
../.env/bin/python manage.py $@''')
    open('src/runserver', 'w').write('''#!/bin/bash
# Helper script for running the local dev server, ensuring
# our virtual environment is used.
#set -e
#script_dir=`dirname $0`
#cd $script_dir
if [ -z "$PORT" ]; then
export PORT=8111
fi
if [ -z "$ROLE" ]; then
export ROLE=dev
fi
. ~/.bash_aliases
./manage runserver localhost:$PORT''')
    open('src/shell', 'w').write(r'''#!/bin/bash
# Creates a local PIP-aware shell.
#set -e
if [ $_ == $0 ]
then
echo "Please source this script. Do not execute."
exit 1
fi
#script_dir=`dirname $0`
#cd $script_dir
. .env/bin/activate
PS1="\u@\h:\W(fab)\$ "''')

    md('media')
    md('static')
    
    os.system('chmod +x src/shell')
    os.system('chmod +x src/manage')
    os.system('chmod +x src/runserver')

    # Create the primary app for containing models/urls/views.
    if not os.path.isdir('src/%s' % args.project_name):
        os.system('cd src; ./manage startapp %s' % (args.project_name,))
        
    os.system('cd src; ./manage syncdb')

if __name__ == "__main__":
    
    parser = argparse.ArgumentParser(
        description=('Initialize and manage a Burlap structured project.'),
    )
    parser.add_argument(
        '-v', '--version',
        help='Show version.',
        action='version',
        version=burlap.__version__,
    )   
    subparsers = parser.add_subparsers(
        dest='action',
    )
    
    skel_parser = subparsers.add_parser(
        SKEL,
        help='Creates a skeleton project.')
    skel_parser.add_argument(
        'project_name',
        type=str,
        help='project name')
    skel_parser.add_argument(
        '--pip-requirements',
        type=str,
        default='Fabric',
        help='The default Python packages to install.')
    skel_parser.add_argument(
        '--virtualenv-dir',
        type=str,
        default='.env',
        help='The default virtualenv directory.')
    skel_parser.add_argument(
        '--roles',
        type=str,
        default='',#dev,prod',
        help='The default roles to populate.')
    skel_parser.add_argument(
        '--components',
        type=str,
        default='',
        help='The default components to enable (e.g. django).')
        
    add_role_parser = subparsers.add_parser(
        ADD_ROLE,
        help='Adds a new role to the project.')
    add_role_parser.add_argument(
        'roles', metavar='roles', type=str, nargs='+',
        help='Names of roles to add.')
                   
    create_satchel_parser = subparsers.add_parser(
        CREATE_SATCHEL,
        help='Adds a new role to the project.')
    create_satchel_parser.add_argument(
        'name', type=str,
        help='Names of the satchel to create.')
        
    args = parser.parse_args()

    if args.action == SKEL:
        assert args.project_name, 'Specify project name.'
        site_name = args.project_name
        
        app_name = args.project_name + '_site'
        
        default_roles = [_ for _ in args.roles.split(',') if _.strip()]
        default_components = [_.strip().lower() for _ in args.components.split(',') if _.strip()]
        
        print('Creating folders...')
        md('roles/all')
        for _role in default_roles:
            md('roles/%s' % _role)
        md('src')
        
        print('Creating roles...')
        open('roles/all/settings.yaml', 'w').write(
            common.render_to_string(
                'burlap/all_settings.yaml.template',
                extra=dict(project_name=args.project_name, site_name=site_name, app_name=app_name)))
        for _role in default_roles:
            open('roles/%s/settings.yaml' % _role, 'w').write(
                common.render_to_string(
                    'burlap/role_settings.yaml.template',
                    extra=dict(project_name=args.project_name, site_name=site_name, role=_role)))
        
        default_packages = args.pip_requirements.split(',')
        if default_packages:
            open('roles/all/pip-requirements.txt', 'w').write('\n'.join(default_packages))
            
        open('roles/all/apt-requirements.txt', 'w').write('')
        
        content = open(fabfile_template, 'r').read()
        content = content.format(project_name=args.project_name)
        open('fabfile.py', 'w').write(content)
        
        print('Initializing local development virtual environment...')
        os.system('virtualenv --no-site-packages %s' % args.virtualenv_dir)
        for package in default_packages:
            os.system('. %s/bin/activate; pip install %s; deactivate' % (args.virtualenv_dir, package))

        # Install burlap dependencies.
        burlap_pip_requirements = os.path.join(os.path.dirname(burlap.__file__), '../requirements.txt')
        print('burlap_pip_requirements:', burlap_pip_requirements)
        assert os.path.exists(burlap_pip_requirements)
        for package in open(burlap_pip_requirements, 'r').readlines():
            if not package.strip():
                continue
            cmd = '%s/bin/pip install %s' % (args.virtualenv_dir, package)
            print('cmd:', cmd)
            assert not os.system(cmd)

        open('setup.bash', 'w').write(common.render_to_string('burlap/setup.bash.template'))
        
        os.system('chmod +x shell')
        
        open('.gitignore', 'w').write(common.render_to_string('burlap/gitignore.template'))
        
        for component in default_components:
            globals()['init_%s' % component](args)
        
        print('='*80)
        print()
        print('Skeleton created for project %s!' % (args.project_name.title(),))
        print()

    elif args.action == ADD_ROLE:
        
        for role in args.roles:
            _role = role.strip().lower()
            fn = 'roles/%s/settings.yaml' % _role
            if os.path.isfile(fn):
                continue
            fn_dir = os.path.split(fn)[0]
            if not os.path.isdir(fn_dir):
                os.makedirs(fn_dir)
            open(fn, 'w').write(
                common.render_to_string(
                    'burlap/role_settings.yaml.template',
                    extra=dict(role=_role)))
            print('Added role %s!' % role)
    
    elif args.action == CREATE_SATCHEL:
        
        name_simple = re.sub(r'[^a-z0-9]+', '', args.name.lower())
        content = common.render_to_string(
            'burlap/satchel.py.template',
            extra=dict(
                name_camelcase=to_camelcase(args.name),
                name_simple=name_simple,
            ))
        if not os.path.isdir('satchels'):
            os.makedirs('satchels')
            os.system('touch satchels/__init__.py')
        satchel_fn = 'satchels/%s.py' % name_simple
        open(satchel_fn, 'w').write(content.strip()+'\n')
        print('Wrote %s.' % satchel_fn)
    
    else:
        raise NotImplementedError, 'Unknown action: %s' % (args.action)
        


"""
System settings
===============
"""
from __future__ import print_function

from fabric.api import hide, run, settings

from burlap.files import file # pylint: disable=redefined-builtin
from burlap.utils import read_lines, run_as_root

is_file = file.is_file

class UnsupportedFamily(Exception):
    """
    Operation not supported on this system family.

    ::

        from burlap.system import UnsupportedFamily, distrib_family

        family = distrib_family()
        if family == 'debian':
            do_some_stuff()
        elif family == 'redhat':
            do_other_stuff()
        else:
            raise UnsupportedFamily(supported=['debian', 'redhat'])

    """

    def __init__(self, supported):
        self.supported = supported
        self.distrib = distrib_id()
        self.family = distrib_family()
        msg = "Unsupported family %s (%s). Supported families: %s" % (self.family, self.distrib, ', '.join(supported))
        super(UnsupportedFamily, self).__init__(msg)


def distrib_id():
    """
    Get the OS distribution ID.

    Returns a string such as ``"Debian"``, ``"Ubuntu"``, ``"RHEL"``,
    ``"CentOS"``, ``"SLES"``, ``"Fedora"``, ``"Arch"``, ``"Gentoo"``,
    ``"SunOS"``...

    Example::

        from burlap.system import distrib_id

        if distrib_id() != 'Debian':
            abort(u"Distribution is not supported")

    """

    with settings(hide('running', 'stdout')):
        kernel = run('uname -s')

        if kernel == 'Linux':
            # lsb_release works on Ubuntu and Debian >= 6.0
            # but is not always included in other distros
            if is_file('/usr/bin/lsb_release'):
                id_ = run('lsb_release --id --short')
                if id in ['arch', 'Archlinux']:  # old IDs used before lsb-release 1.4-14
                    id_ = 'Arch'
                return id_
            else:
                if is_file('/etc/debian_version'):
                    return "Debian"
                elif is_file('/etc/fedora-release'):
                    return "Fedora"
                elif is_file('/etc/arch-release'):
                    return "Arch"
                elif is_file('/etc/redhat-release'):
                    release = run('cat /etc/redhat-release')
                    if release.startswith('Red Hat Enterprise Linux'):
                        return "RHEL"
                    elif release.startswith('CentOS'):
                        return "CentOS"
                    elif release.startswith('Scientific Linux'):
                        return "SLES"
                elif is_file('/etc/gentoo-release'):
                    return "Gentoo"
        elif kernel == "SunOS":
            return "SunOS"


def distrib_release():
    """
    Get the release number of the distribution.

    Example::

        from burlap.system import distrib_id, distrib_release

        if distrib_id() == 'CentOS' and distrib_release() == '6.1':
            print(u"CentOS 6.2 has been released. Please upgrade.")

    """

    with settings(hide('running', 'stdout')):

        kernel = run('uname -s')

        if kernel == 'Linux':
            return run('lsb_release -r --short')

        elif kernel == 'SunOS':
            return run('uname -v')


def distrib_codename():
    """
    Get the codename of the Linux distribution.

    Example::

        from burlap.deb import distrib_codename

        if distrib_codename() == 'precise':
            print(u"Ubuntu 12.04 LTS detected")

    """
    with settings(hide('running', 'stdout')):
        return run('lsb_release --codename --short')


def distrib_desc():
    """
    Get the description of the Linux distribution.

    For example: ``Debian GNU/Linux 6.0.7 (squeeze)``.
    """
    with settings(hide('running', 'stdout')):
        if not is_file('/etc/redhat-release'):
            return run('lsb_release --desc --short')
        return run('cat /etc/redhat-release')


def distrib_family():
    """
    Get the distribution family.

    Returns one of ``debian``, ``redhat``, ``arch``, ``gentoo``,
    ``sun``, ``other``.
    """
    distrib = distrib_id()
    if distrib in ['Debian', 'Ubuntu', 'LinuxMint', 'elementary OS']:
        return 'debian'
    elif distrib in ['RHEL', 'CentOS', 'SLES', 'Fedora']:
        return 'redhat'
    elif distrib in ['SunOS']:
        return 'sun'
    elif distrib in ['Gentoo']:
        return 'gentoo'
    elif distrib in ['Arch', 'ManjaroLinux']:
        return 'arch'
    else:
        return 'other'


def get_hostname():
    """
    Get the fully qualified hostname.
    """
    with settings(hide('running', 'stdout')):
        return run('hostname --fqdn')


def set_hostname(hostname, persist=True):
    """
    Set the hostname.
    """
    run_as_root('hostname %s' % hostname)
    if persist:
        run_as_root('echo %s >/etc/hostname' % hostname)


def get_sysctl(key):
    """
    Get a kernel parameter.

    Example::

        from burlap.system import get_sysctl

        print("Max number of open files:", get_sysctl('fs.file-max')

    """
    with settings(hide('running', 'stdout')):
        return run_as_root('/sbin/sysctl -n -e %(key)s' % locals())


def set_sysctl(key, value):
    """
    Set a kernel parameter.

    Example::

        import burlap

        # Protect from SYN flooding attack
        burlap.system.set_sysctl('net.ipv4.tcp_syncookies', 1)

    """
    run_as_root('/sbin/sysctl -n -e -w %(key)s=%(value)s' % locals())


def supported_locales():
    """
    Gets the list of supported locales.

    Each locale is returned as a ``(locale, charset)`` tuple.
    """
    family = distrib_family()
    if family == 'debian':
        return _parse_locales('/usr/share/i18n/SUPPORTED')
    elif family == 'arch':
        return _parse_locales('/etc/locale.gen')
    elif family == 'redhat':
        return _supported_locales_redhat()
    else:
        raise UnsupportedFamily(supported=['debian', 'arch', 'redhat'])


def _parse_locales(path):
    lines = read_lines(path)
    return list(_split_on_spaces(_strip(_remove_comments(lines))))


def _split_on_spaces(lines):
    return (line.split(' ') for line in lines)


def _strip(lines):
    return (line.strip() for line in lines)


def _remove_comments(lines):
    return (line for line in lines if not line.startswith('#'))


def _supported_locales_redhat():
    res = run('/usr/bin/locale -a')
    return [(locale, None) for locale in res.splitlines()]


def get_arch():
    """
    Get the CPU architecture.

    Example::

        from burlap.system import get_arch

        if get_arch() == 'x86_64':
            print(u"Running on a 64-bit Intel/AMD system")

    """
    with settings(hide('running', 'stdout')):
        arch = run('uname -m')
        return arch


def cpus():
    """
    Get the number of CPU cores.

    Example::

        from burlap.system import cpus

        nb_workers = 2 * cpus() + 1

    """
    with settings(hide('running', 'stdout')):
        res = run('python -c "import multiprocessing; '
                  'print(multiprocessing.cpu_count())"')
        return int(res)


def using_systemd():
    """
    Return True if using systemd

    Example::

        from burlap.system import use_systemd

        if using_systemd():
            # do stuff with burlap.systemd ...
            pass

    """
    return run('which systemctl', quiet=True).succeeded


def time():
    """
    Return the current time in seconds since the Epoch.

    Same as :py:func:`time.time()`

    """

    with settings(hide('running', 'stdout')):
        return int(run('date +%s'))



"""
MySQL users and databases
=========================

This module provides tools for creating MySQL users and databases.

"""
from __future__ import print_function

import os
import re
from pipes import quote

from fabric.api import env, hide, puts, run, settings, runs_once
from fabric.colors import red, green

from burlap import Satchel
from burlap.constants import *
from burlap.db import DatabaseSatchel
from burlap.decorators import task
from burlap.utils import run_as_root

MYSQLD_SAFE = 'mysqld_safe'
MYSQLADMIN = 'mysqladmin'
DPKG = 'dpkg'

class MySQLSatchel(DatabaseSatchel):
    
    name = 'mysql'
    
    @property
    def packager_system_packages(self):
        return {
            FEDORA: ['mysql-server'],
            (UBUNTU, '12.04'): ['mysql-server', 'libmysqlclient-dev'],
            (UBUNTU, '14.04'): ['mysql-server-5.6', 'libmysqlclient-dev'],
            (UBUNTU, '16.04'): ['mysql-server', 'libmysqlclient-dev'],
        }
    
    def set_defaults(self):
        super(MySQLSatchel, self).set_defaults()
    
        # You want this to be large, and set in both the client and server.
        # Otherwise, MySQL may silently truncate database dumps, leading to much
        # frustration.
        self.env.max_allowed_packet = 524288000 # 500M
        
        self.env.net_buffer_length = 1000000
        self.env.conf = '/etc/mysql/my.cnf' # /etc/my.cnf on fedora
        
        self.env.dump_command = 'mysqldump --opt --compress --max_allowed_packet={max_allowed_packet} ' \
            '--force --single-transaction --quick --user {db_user} ' \
            '--password={db_password} -h {db_host} {db_name} | gzip > {dump_fn}'
        
        self.env.load_command = 'gunzip < {remote_dump_fn} | mysql -u {db_root_username} ' \
            '--password={db_root_password} --host={db_host} -D {db_name}'
        
        self.env.preload_commands = []
        self.env.character_set = 'utf8'
        self.env.collate = 'utf8_general_ci'
        self.env.port = 3306
        self.env.root_username = 'root'
        self.env.root_password = None
        self.env.custom_mycnf = False
        
        self.env.assumed_version = '5.7'

        self.env.service_commands = {
            START:{
                UBUNTU: 'service mysql start',
            },
            STOP:{
                UBUNTU: 'service mysql stop',
            },
            ENABLE:{
                UBUNTU: 'update-rc.d mysql defaults',
            },
            DISABLE:{
                UBUNTU: 'update-rc.d -f mysql remove',
            },
            RESTART:{
                UBUNTU: 'service mysql restart',
            },
            STATUS:{
                UBUNTU: 'service mysql status',
            },
        }
        
    @task
    def execute(self, sql, name='default', site=None, **kwargs):
        use_sudo = int(kwargs.pop('use_sudo', 0))
        r = self.database_renderer(name=name, site=site)
        r.env.user = kwargs.pop('user', r.env.db_root_username)
        r.env.password = kwargs.pop('password', r.env.db_root_password)
        r.env.sql = sql
        cmd = "mysql --user={user} -p'{db_root_password}' --execute='{sql}'"
        if use_sudo:
            r.sudo(cmd)
        else:
            r.run(cmd)

    @task
    def execute_file(self, filename, name='default', site=None, **kwargs):
        r = self.database_renderer(name=name, site=site)
        r.env.user = kwargs.pop('user', r.env.db_root_username)
        r.env.password = kwargs.pop('password', r.env.db_root_password)
        r.env.filename = filename
        r.run("mysql --user={user} -p'{db_root_password}' {db_name} < {filename}")

    @task
    def set_collation(self, name=None, site=None):
        r = self.database_renderer(name=name, site=site)
        r.run("mysql -v -h {db_host} -u {db_root_username} -p'{db_root_password}' "
            "--execute='ALTER DATABASE {db_name} CHARACTER SET {character_set} COLLATE {collate};'")
    
    @task
    def set_collation_all(self, name=None, site=None):
        for site in self.genv.available_sites:
            self.set_collation(name=name, site=site)
    
    @task
    def set_max_packet_size(self, name=None, site=None):
        r = self.database_renderer(name=name, site=site)
        r.run(
            ('mysql -v -h {db_host} -D {db_name} -u {db_root_username} '
            '-p"{db_root_password}" --execute="SET global '
            'net_buffer_length={net_buffer_length}; SET global '
            'max_allowed_packet={max_allowed_packet};"') % env)
    
    def packager_pre_configure(self):
        """
        Called before packager.configure is run.
        """
        self.prep_root_password()
        
    @task
    def prep_root_password(self, password=None, **kwargs):
        """
        Enters the root password prompt entries into the debconf cache
        so we can set them without user interaction.
        
        We keep this process separate from set_root_password() because we also need to do
        this before installing the base MySQL package, because that will also prompt the user
        for a root login.
        """
        r = self.database_renderer(**kwargs)
        r.env.root_password = password or r.genv.get('db_root_password')
        r.sudo("DEBIAN_FRONTEND=noninteractive dpkg --configure -a")
        r.sudo("debconf-set-selections <<< 'mysql-server mysql-server/root_password password {root_password}'")
        r.sudo("debconf-set-selections <<< 'mysql-server mysql-server/root_password_again password {root_password}'")
    
    @task
    def get_mysql_version(self):
        return (self.run("dpkg --list | grep -oP '(?<=mysql-server-)([0-9.]+)'") or self.env.assumed_version).split('\n')[0].strip()
    
    @task
    def assert_mysql_stopped(self):
        with self.settings(warn_only=True):
            ret = (self.run('ps aux |grep -i mysql|grep -v grep|grep -v vagrant') or '').strip()
        assert not ret
    
    @task
    def set_root_password(self, password=None, method=None, **kwargs):
        method = method or MYSQLD_SAFE#|'mysqladmin'#|'mysqld_safe'|'dpkg'
        v = self.get_mysql_version()
        v = tuple(map(int, v.split('.')))
        self.vprint('mysql version:', v)
        if method == MYSQLADMIN:
            r = self.database_renderer(**kwargs)
            r.env.root_password = password or r.env.db_root_password
            r.sudo('mysqladmin -u root password {root_password}')
        elif method == DPKG:
            #TODO:fix? This no longer prompts to set root password with >= 5.7.
            self.prep_root_password(**kwargs)
            r = self.database_renderer(**kwargs)
            r.sudo("dpkg-reconfigure -fnoninteractive `dpkg --list | egrep -o 'mysql-server-([0-9.]+)'`")
        elif method == MYSQLD_SAFE:
            #TODO:fix? unreliable?
            #https://dev.mysql.com/doc/refman/5.7/en/resetting-permissions.html
            r = self.database_renderer(**kwargs)
            r.env.root_password = password or r.env.db_root_password
            
            # Confirm server stopped.
            self.stop()
            self.assert_mysql_stopped()
            #r.sudo('mysqladmin shutdown')
            
            r.sudo('mkdir -p /var/run/mysqld')
            r.sudo('chown mysql /var/run/mysqld')
            
            # Note we have to use pty=False here, otherwise, even with nohup, the process gets killed as soon as the sudo call exits.
            # http://stackoverflow.com/a/27600071/247542
            r.sudo('nohup mysqld_safe --skip-grant-tables &> /tmp/mysqld_safe.log < /dev/null &', pty=False)
            
            running = False
            for _wait in range(10):
                r.run('sleep 1')
                with self.settings(warn_only=True):
                    ret = (r.run('ps aux|grep -i mysql|grep -v grep|grep -v vagrant') or '').strip()
                if len(ret):
                    running = True
                    break
            if not running and not self.dryrun:
                raise Exception('Could not launch mysqld_safe.')
            r.run('sleep 5')
            # Work in Ubuntu 16/MySQL 5.7 but not Ubuntu 14/MySQL 5.6?
            with settings(warn_only=True):
                r.run("mysql -uroot --execute=\""
                    "use mysql; "
                    "update user set authentication_string=PASSWORD('{root_password}') where User='root'; "
                    "flush privileges;\"")
            # Work in Ubuntu 14/MySQL 5.6 but not Ubuntu 16/MySQL 5.7?
            with settings(warn_only=True):
                r.sudo('mysql --execute="USE mysql; SET PASSWORD FOR \'root\'@\'localhost\' = PASSWORD(\'{root_password}\'); FLUSH PRIVILEGES;"')
            
            # Signal server to stop.
            # Note, `sudo service mysql stop` and `sudo /etc/init.d/mysql stop` and `mysqladmin shutdown` don't seem to work with mysqld_safe.
            r.sudo("[ -f /var/run/mysqld/mysqld.pid ] && kill `sudo cat /var/run/mysqld/mysqld.pid` || true")
            
            # Confirm server stopped.
            r.run('sleep 10')
            self.assert_mysql_stopped()
            
            self.start()
        else:
            raise NotImplementedError('Unknowne method: %s' % method)
        
    @task
    def dumpload(self, site=None, role=None):
        """
        Dumps and loads a database snapshot simultaneously.
        Requires that the destination server has direct database access
        to the source server.
        
        This is better than a serial dump+load when:
        1. The network connection is reliable.
        2. You don't need to save the dump file.
        
        The benefits of this over a dump+load are:
        1. Usually runs faster, since the load and dump happen in parallel.
        2. Usually takes up less disk space since no separate dump file is
            downloaded.
        """
        raise NotImplementedError
        
    @task
    def drop_views(self, name=None, site=None):
        """
        Drops all views.
        """
        
        r = self.database_renderer
            
        result = r.sudo("mysql --batch -v -h {db_host} "
            #"-u {db_root_username} -p'{db_root_password}' "
            "-u {db_user} -p'{db_password}' "
            "--execute=\"SELECT GROUP_CONCAT(CONCAT(TABLE_SCHEMA,'.',table_name) SEPARATOR ', ') AS views "
            "FROM INFORMATION_SCHEMA.views WHERE TABLE_SCHEMA = '{db_name}' ORDER BY table_name DESC;\"")
        result = re.findall(
            r'^views[\s\t\r\n]+(.*)',
            result,
            flags=re.IGNORECASE|re.DOTALL|re.MULTILINE)
        if not result:
            return
        r.env.db_view_list = result[0]
        #cmd = ("mysql -v -h {db_host} -u {db_root_username} -p'{db_root_password}' " \
        r.sudo("mysql -v -h {db_host} -u {db_user} -p'{db_password}' " \
            "--execute=\"DROP VIEW {db_view_list} CASCADE;\"")
        
    @task
    @runs_once
    def exists(self, **kwargs):
        """
        Returns true if a database with the given name exists. False otherwise.
        """
        
        name = kwargs.pop('name', 'default')
        site = kwargs.pop('site', None)
        
        r = self.database_renderer(name=name, site=site)
        
        ret = r.run('mysql -h {db_host} -u {db_root_username} '\
            '-p"{db_root_password}" -N -B -e "SELECT IF(\'{db_name}\''\
            ' IN(SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA), '\
            '\'exists\', \'notexists\') AS found;"')
            
        if ret is not None:
            ret = 'notexists' not in (ret or 'notexists')
        
        if ret is not None:
            msg = '%s database on site %s %s exist.' \
                % (name.title(), env.SITE, 'DOES' if ret else 'DOES NOT')
            if ret:
                print(green(msg))
            else:
                print(red(msg))
            return ret

    @task
    @runs_once
    def create(self, **kwargs):
        
        name = kwargs.pop('name', 'default')
        site = kwargs.pop('site', None)
        drop = int(kwargs.pop('drop', 0))
        #post_process = int(kwargs.pop('post_process', 0))
        
        r = self.database_renderer(name=name, site=site)
        
        # Do nothing if we're not dropping and the database already exists.
        print('Checking to see if database already exists...')
        if self.exists(name=name, site=site) and not drop:
            print('Database already exists. Aborting creation. '\
                'Use drop=1 to override.')
            return
            
        r.env.db_drop_flag = '--drop' if drop else ''
        
        if int(drop):
            r.sudo("mysql -v -h {db_host} -u {db_root_username} -p'{db_root_password}' "\
                "--execute='DROP DATABASE IF EXISTS {db_name}'")
            
        r.sudo("mysqladmin -h {db_host} -u {db_root_username} -p'{db_root_password}' create {db_name}")
 
        self.set_collation(name=name, site=site)
            
        # Create user.
        with self.settings(warn_only=True):
            r.run("mysql -v -h {db_host} -u {db_root_username} -p'{db_root_password}' "\
                "--execute=\"GRANT USAGE ON *.* TO {db_user}@'%%'; DROP USER {db_user}@'%%';\"")
        
        # Grant user access to the database.
        r.run("mysql -v -h {db_host} -u {db_root_username} "\
            "-p'{db_root_password}' --execute=\"GRANT ALL PRIVILEGES "\
            "ON {db_name}.* TO {db_user}@'%%' IDENTIFIED BY "\
            "'{db_password}'; FLUSH PRIVILEGES;\"")
        #TODO:why is this necessary? why doesn't the user@% pattern above give
        #localhost access?!
        r.run("mysql -v -h {db_host} -u {db_root_username} "\
            "-p'{db_root_password}' --execute=\"GRANT ALL PRIVILEGES "\
            "ON {db_name}.* TO {db_user}@{db_host} IDENTIFIED BY "\
            "'{db_password}'; FLUSH PRIVILEGES;\"")
        
    @task
    @runs_once
    def load(self, dump_fn='', prep_only=0, force_upload=0, from_local=0, name=None, site=None, dest_dir=None):
        """
        Restores a database snapshot onto the target database server.
        
        If prep_only=1, commands for preparing the load will be generated,
        but not the command to finally load the snapshot.
        """
        
        r = self.database_renderer(name=name, site=site)
        r.pc('Loading database snapshot.')
        
        # Render the snapshot filename.
        r.env.dump_fn = self.get_default_db_fn(fn_template=dump_fn, dest_dir=dest_dir).strip()
        
        from_local = int(from_local)
        
        prep_only = int(prep_only)
        
        missing_local_dump_error = r.format(
            "Database dump file {dump_fn} does not exist."
        )
        
        # Copy snapshot file to target.
        if r.genv.is_local:
            r.env.remote_dump_fn = dump_fn
        else:
            r.env.remote_dump_fn = '/tmp/' + os.path.split(r.env.dump_fn)[-1]
        
        if not prep_only:
            if int(force_upload) or (not r.genv.is_local and not r.file_exists(r.env.remote_dump_fn)):
                if not self.dryrun:
                    assert os.path.isfile(r.env.dump_fn), \
                        missing_local_dump_error
                if self.verbose:
                    print('Uploading MySQL database snapshot...')
                r.put(
                    local_path=r.env.dump_fn,
                    remote_path=r.env.remote_dump_fn)
        
        if r.genv.is_local and not prep_only and not self.dryrun:
            assert os.path.isfile(r.env.dump_fn), \
                missing_local_dump_error
        
        
        # Drop the database if it's there.
        r.run("mysql -v -h {db_host} -u {db_root_username} -p'{db_root_password}' "
            "--execute='DROP DATABASE IF EXISTS {db_name}'")
        
        # Now, create the database.
        r.run("mysqladmin -h {db_host} -u {db_root_username} -p'{db_root_password}' create {db_name}")
        
        # Create user
        with settings(warn_only=True):
            r.run("mysql -v -h {db_host} -u {db_root_username} -p'{db_root_password}' "
                "--execute=\"CREATE USER '{db_user}'@'%%' IDENTIFIED BY '{db_password}'; "
                "GRANT ALL PRIVILEGES ON *.* TO '{db_user}'@'%%' WITH GRANT OPTION; "
                "FLUSH PRIVILEGES;\"")
        self.set_collation(name=name, site=site)
        
        self.set_max_packet_size(name=name, site=site)
        
        # Run any server-specific commands (e.g. to setup permissions) before
        # we load the data.
        for command in r.env.preload_commands:
            r.run(command)
        
        # Restore the database content from the dump file.
        if not prep_only:
            r.run(r.env.load_command)
        
        self.set_collation(name=name, site=site)

    @task
    def shell(self, name='default', site=None, **kwargs):
        """
        Opens a SQL shell to the given database, assuming the configured database
        and user supports this feature.
        """
        r = self.database_renderer(name=name, site=site)
        r.run('/bin/bash -i -c "mysql -u {db_user} -p\'{db_password}\' -h {db_host} {db_name}"')

    @task(precursors=['packager', 'user'])
    def configure(self, do_packages=0, name='default', site=None):

        r = self.database_renderer(name=name, site=site)
        
        if int(do_packages):
            self.prep_root_password()
            self.install_packages()
            
        self.set_root_password()
        
        if r.env.custom_mycnf:
            fn = r.render_to_file('mysql/my.template.cnf', extra=r.env)
            r.put(
                local_path=fn,
                remote_path=r.env.conf,
                use_sudo=True,
            )
        
        if r.env.allow_remote_connections:
            
            # Enable remote connections.
            r.sudo("sed -i 's/127.0.0.1/0.0.0.0/g' {conf}")
            
            # Enable root logins from remote connections.
            r.sudo('mysql -u {db_root_username} -p"{db_root_password}" '
                '--execute="USE mysql; '
                'GRANT ALL ON *.* to {db_root_username}@\'%%\' IDENTIFIED BY \'{db_root_password}\'; '
                'FLUSH PRIVILEGES;"')
            
            self.restart()

class MySQLClientSatchel(Satchel):

    name = 'mysqlclient'
    
    @property
    def packager_system_packages(self):
        return {
            (UBUNTU, '12.04'): ['libmysqlclient-dev', 'mysql-client'],
            (UBUNTU, '14.04'): ['libmysqlclient-dev', 'mysql-client'],
            (UBUNTU, '16.04'): ['libmysqlclient-dev', 'mysql-client'],
        }
        
    @task(precursors=['packager'])
    def configure(self, *args, **kwargs):
        pass

mysql = MySQLSatchel()
mysqlclient = MySQLClientSatchel()

def query(query, use_sudo=True, **kwargs):
    """
    Run a MySQL query.
    """
    func = use_sudo and run_as_root or run

    user = kwargs.get('mysql_user') or env.get('mysql_user')
    password = kwargs.get('mysql_password') or env.get('mysql_password')

    options = [
        '--batch',
        '--raw',
        '--skip-column-names',
    ]
    if user:
        options.append('--user=%s' % quote(user))
    if password:
        options.append('--password=%s' % quote(password))
    options = ' '.join(options)

    return func('mysql %(options)s --execute=%(query)s' % {
        'options': options,
        'query': quote(query),
    })


def user_exists(name, host='localhost', **kwargs):
    """
    Check if a MySQL user exists.
    """
    with settings(hide('running', 'stdout', 'stderr', 'warnings'), warn_only=True):
        res = query("""
            use mysql;
            SELECT COUNT(*) FROM user
                WHERE User = '%(name)s' AND Host = '%(host)s';
            """ % {
                'name': name,
                'host': host,
            }, **kwargs)
    return res.succeeded and (int(res) == 1)


def create_user(name, password, host='localhost', **kwargs):
    """
    Create a MySQL user.

    Example::

        import burlap

        # Create DB user if it does not exist
        if not burlap.mysql.user_exists('dbuser'):
            burlap.mysql.create_user('dbuser', password='somerandomstring')

    """
    with settings(hide('running')):
        query("CREATE USER '%(name)s'@'%(host)s' IDENTIFIED BY '%(password)s';" % {
            'name': name,
            'password': password,
            'host': host
        }, **kwargs)
    puts("Created MySQL user '%s'." % name)


def database_exists(name, **kwargs):
    """
    Check if a MySQL database exists.
    """
    with settings(hide('running', 'stdout', 'stderr', 'warnings'), warn_only=True):
        res = query("SHOW DATABASES LIKE '%(name)s';" % {
            'name': name
        }, **kwargs)

    return res.succeeded and (res == name)


def create_database(name, owner=None, owner_host='localhost', charset='utf8',
                    collate='utf8_general_ci', **kwargs):
    """
    Create a MySQL database.

    Example::

        import burlap

        # Create DB if it does not exist
        if not burlap.mysql.database_exists('myapp'):
            burlap.mysql.create_database('myapp', owner='dbuser')

    """
    with settings(hide('running')):

        query("CREATE DATABASE %(name)s CHARACTER SET %(charset)s COLLATE %(collate)s;" % {
            'name': name,
            'charset': charset,
            'collate': collate
        }, **kwargs)

        if owner:
            query("GRANT ALL PRIVILEGES ON %(name)s.* TO '%(owner)s'@'%(owner_host)s' WITH GRANT OPTION;" % {
                'name': name,
                'owner': owner,
                'owner_host': owner_host
            }, **kwargs)

    puts("Created MySQL database '%s'." % name)



import hashlib

from fabric.api import hide, run, settings

import burlap


def test_md5sum_empty_file():
    try:
        run('touch f1')
        expected_hash = hashlib.md5('').hexdigest()
        assert burlap.files.file.md5sum('f1') == expected_hash
    finally:
        run('rm -f f1')


def test_md5sum():
    try:
        run('echo -n hello > f2')
        expected_hash = hashlib.md5('hello').hexdigest()
        assert burlap.files.file.md5sum('f2') == expected_hash
    finally:
        run('rm -f f2')


def test_md5sum_not_existing_file():
    with settings(hide('warnings')):
        assert burlap.files.file.md5sum('doesnotexist') is None



"""
Wrapper around the Motion service.

http://www.lavrsen.dk/foswiki/bin/view/Motion/WebHome
"""
from __future__ import print_function

import feedparser

from burlap import Satchel
from burlap.constants import *
from burlap.decorators import task

class SeleniumSatchel(Satchel):
    """
    Management commands for the Selenium browser automation and testing tool.
    
    http://www.seleniumhq.org/
    """

    name = 'selenium'

    def set_defaults(self):        
        # See https://github.com/mozilla/geckodriver/releases for other versions and architectures.
        self.env.geckodriver_version = '0.13.0'
        self.env.geckodriver_arch = 'linux64'
        self.env.geckodriver_url_template = \
            'https://github.com/mozilla/geckodriver/releases/download/' \
            'v{geckodriver_version}/geckodriver-v{geckodriver_version}-{geckodriver_arch}.tar.gz'
        self.env.geckodriver_install_bin_path = '/usr/local/bin'
        self.env.geckodriver_bin_name = 'geckodriver'
        self.env.geckodriver_fingerprint_path = '/usr/local/lib/geckodriver/fingerprint.txt'

    @property
    def geckodriver_path(self):
        r = self.local_renderer
        return r.format('{geckodriver_install_bin_path}/{geckodriver_bin_name}')

    @task
    def install_geckodriver(self):
        r = self.local_renderer
        r.run(
            'cd /tmp; '
            'wget -O geckodriver.tar.gz {geckodriver_url_template}; '
            'tar -xvzf geckodriver.tar.gz')
        r.sudo('mv /tmp/{geckodriver_bin_name} {geckodriver_install_bin_path}')

    @task
    def uninstall_geckodriver(self):
        r = self.local_renderer
        r.sudo('rm -f {geckodriver_install_bin_path}/{geckodriver_bin_name}')

    @task
    def check_for_change(self):
        """
        Determines if a new release has been made.
        """
        r = self.local_renderer
        lm = self.last_manifest
        last_fingerprint = lm.fingerprint
        current_fingerprint = self.get_fingerprint()
        print('last_fingerprint:', last_fingerprint)
        print('current_fingerprint:', current_fingerprint)
        if last_fingerprint != current_fingerprint:
            print('A new release is available.')
        else:
            print('No updates found.')
        
    @task
    def get_fingerprint(self):
        fingerprint = feedparser.parse('https://github.com/mozilla/geckodriver/tags.atom')['entries'][0]['link']
        self.vprint('fingerprint:', fingerprint)
        return fingerprint

    def record_manifest(self):
        """
        Called after a deployment to record any data necessary to detect changes
        for a future deployment.
        """
        manifest = super(SeleniumSatchel, self).record_manifest()
        manifest['fingerprint'] = self.get_fingerprint()
        return manifest

    @task(precursors=['packager'])
    def configure(self):
        if self.env.enabled:
            self.install_geckodriver()
        else:
            self.uninstall_geckodriver()

selenium = SeleniumSatchel()



"""
PostgreSQL users and databases
==============================

This module provides tools for creating PostgreSQL users and databases.

"""
from __future__ import print_function

import os

from fabric.api import cd, hide, sudo, settings, runs_once

from burlap import Satchel
from burlap.constants import *
from burlap.db import DatabaseSatchel
from burlap.decorators import task

POSTGIS = 'postgis'
POSTGRESQL = 'postgresql'

def _run_as_pg(command):
    """
    Run command as 'postgres' user
    """
    with cd('~postgres'):
        return sudo(command, user='postgres')


def user_exists(name):
    """
    Check if a PostgreSQL user exists.
    """
    with settings(hide('running', 'stdout', 'stderr', 'warnings'),
                  warn_only=True):
        res = _run_as_pg('''psql -t -A -c "SELECT COUNT(*) FROM pg_user WHERE usename = '%(name)s';"''' % locals())
    return (res == "1")


def create_user(name, password, superuser=False, createdb=False,
                createrole=False, inherit=True, login=True,
                connection_limit=None, encrypted_password=False):
    """
    Create a PostgreSQL user.

    Example::

        import burlap

        # Create DB user if it does not exist
        if not burlap.postgres.user_exists('dbuser'):
            burlap.postgres.create_user('dbuser', password='somerandomstring')

        # Create DB user with custom options
        burlap.postgres.create_user('dbuser2', password='s3cr3t',
            createdb=True, createrole=True, connection_limit=20)

    """
    options = [
        'SUPERUSER' if superuser else 'NOSUPERUSER',
        'CREATEDB' if createdb else 'NOCREATEDB',
        'CREATEROLE' if createrole else 'NOCREATEROLE',
        'INHERIT' if inherit else 'NOINHERIT',
        'LOGIN' if login else 'NOLOGIN',
    ]
    if connection_limit is not None:
        options.append('CONNECTION LIMIT %d' % connection_limit)
    password_type = 'ENCRYPTED' if encrypted_password else 'UNENCRYPTED'
    options.append("%s PASSWORD '%s'" % (password_type, password))
    options = ' '.join(options)
    _run_as_pg('''psql -c "CREATE USER %(name)s %(options)s;"''' % locals())


def drop_user(name):
    """
    Drop a PostgreSQL user.

    Example::

        import burlap

        # Remove DB user if it exists
        if burlap.postgres.user_exists('dbuser'):
            burlap.postgres.drop_user('dbuser')

    """
    _run_as_pg('''psql -c "DROP USER %(name)s;"''' % locals())


def database_exists(name):
    """
    Check if a PostgreSQL database exists.
    """
    with settings(hide('running', 'stdout', 'stderr', 'warnings'),
                  warn_only=True):
        return _run_as_pg('''psql -d %(name)s -c ""''' % locals()).succeeded


def create_database(name, owner, template='template0', encoding='UTF8',
                    locale='en_US.UTF-8'):
    """
    Create a PostgreSQL database.

    Example::

        import burlap

        # Create DB if it does not exist
        if not burlap.postgres.database_exists('myapp'):
            burlap.postgres.create_database('myapp', owner='dbuser')

    """
    _run_as_pg('''createdb --owner %(owner)s --template %(template)s \
                  --encoding=%(encoding)s --lc-ctype=%(locale)s \
                  --lc-collate=%(locale)s %(name)s''' % locals())


def drop_database(name):
    """
    Delete a PostgreSQL database.

    Example::

        import burlap

        # Remove DB if it exists
        if burlap.postgres.database_exists('myapp'):
            burlap.postgres.drop_database('myapp')

    """
    _run_as_pg('''dropdb %(name)s''' % locals())


def create_schema(name, database, owner=None):
    """
    Create a schema within a database.
    """
    if owner:
        _run_as_pg('''psql %(database)s -c "CREATE SCHEMA %(name)s AUTHORIZATION %(owner)s"''' % locals())
    else:
        _run_as_pg('''psql %(database)s -c "CREATE SCHEMA %(name)s"''' % locals())

class PostgreSQLSatchel(DatabaseSatchel):
    """
    Represents a PostgreSQL server.
    """
    
    name = 'postgresql'
    
    @property
    def packager_system_packages(self):
        return {
            (UBUNTU, '12.04'): ['postgresql-9.1'],
            (UBUNTU, '14.04'): ['postgresql-9.3'],
            (UBUNTU, '16.04'): ['postgresql-9.5'],
        }
    
    def set_defaults(self):
        super(PostgreSQLSatchel, self).set_defaults()
        
        # Note, if you use gzip, you can't use parallel restore.
        #self.env.dump_command = 'time pg_dump -c -U {db_user} --no-password --blobs --format=c --schema=public --host={db_host} {db_name} > {dump_fn}'
        self.env.dump_command = 'time pg_dump -c -U {db_user} --no-password --blobs --format=c --schema=public --host={db_host} {db_name} | gzip -c > {dump_fn}'
        self.env.dump_fn_template = '{dump_dest_dir}/db_{db_type}_{SITE}_{ROLE}_{db_name}_$(date +%Y%m%d).sql.gz'
        
        #self.env.load_command = 'gunzip < {remote_dump_fn} | pg_restore --jobs=8 -U {db_root_username} --format=c --create --dbname={db_name}'
        self.env.load_command = 'gunzip < {remote_dump_fn} | pg_restore -U {db_root_username} --format=c --create --dbname={db_name}'
        
        self.env.createlangs = ['plpgsql'] # plpythonu
        self.env.postgres_user = 'postgres'
        self.env.encoding = 'UTF8'
        self.env.custom_load_cmd = ''
        self.env.port = 5432
        self.env.pgpass_path = '~/.pgpass'
        self.env.pgpass_chmod = 600
        self.env.force_version = None
        self.env.version_command = '`psql --version | grep -o -E "[0-9]+.[0-9]+"`'
        self.env.engine = POSTGRESQL # 'postgresql' | postgis

        self.env.apt_repo_enabled = False
        
        # Populated from https://www.postgresql.org/download/linux/ubuntu/
        self.env.apt_repo = None
        
        self.env.apt_key = 'https://www.postgresql.org/media/keys/ACCC4CF8.asc'

        self.env.service_commands = {
            START:{
                UBUNTU: 'service postgresql start',
            },
            STOP:{
                UBUNTU: 'service postgresql stop',
            },
            ENABLE:{
                UBUNTU: 'update-rc.d postgresql defaults',
            },
            DISABLE:{
                UBUNTU: 'update-rc.d -f postgresql remove',
            },
            RESTART:{
                UBUNTU: 'service postgresql restart',
            },
            STATUS:{
                UBUNTU: 'service postgresql status',
            },
        }

    @task
    def write_pgpass(self, name=None, site=None, use_sudo=0, root=0):
        """
        Write the file used to store login credentials for PostgreSQL.
        """
        
        r = self.database_renderer(name=name, site=site)
        
        root = int(root)
        
        use_sudo = int(use_sudo)
        
        r.run('touch {pgpass_path}')
        r.sudo('chmod {pgpass_chmod} {pgpass_path}')
        
        if root:
            r.env.shell_username = r.env.db_root_username
            r.env.shell_password = r.env.db_root_password
        else:
            r.env.shell_username = r.env.db_user
            r.env.shell_password = r.env.db_password
        
        r.append(
            '{db_host}:{port}:*:{shell_username}:{shell_password}',
            r.env.pgpass_path,
            use_sudo=use_sudo)
            
    @task
    def dumpload(self, site=None, role=None):
        """
        Dumps and loads a database snapshot simultaneously.
        Requires that the destination server has direct database access
        to the source server.
        
        This is better than a serial dump+load when:
        1. The network connection is reliable.
        2. You don't need to save the dump file.
        
        The benefits of this over a dump+load are:
        1. Usually runs faster, since the load and dump happen in parallel.
        2. Usually takes up less disk space since no separate dump file is
            downloaded.
        """
        r = self.database_renderer(site=site, role=role)
        r.run('pg_dump -c --host={host_string} --username={db_user} '
            '--blobs --format=c {db_name} -n public | '
            'pg_restore -U {db_postgresql_postgres_user} --create '
            '--dbname={db_name}')
        
    @task
    def drop_views(self, name=None, site=None):
        """
        Drops all views.
        """
        raise NotImplementedError
    #        SELECT 'DROP VIEW ' || table_name || ';'
    #        FROM information_schema.views
    #        WHERE table_schema NOT IN ('pg_catalog', 'information_schema')
    #        AND table_name !~ '^pg_';
            # http://stackoverflow.com/questions/13643831/drop-all-views-postgresql
    #        DO$$
    #        BEGIN
    #        
    #        EXECUTE (
    #           SELECT string_agg('DROP VIEW ' || t.oid::regclass || ';', ' ')  -- CASCADE?
    #           FROM   pg_class t
    #           JOIN   pg_namespace n ON n.oid = t.relnamespace
    #           WHERE  t.relkind = 'v'
    #           AND    n.nspname = 'my_messed_up_schema'
    #           );
    #        
    #        END
    #        $$

    @task
    def exists(self, name='default', site=None):
        """
        Returns true if a database with the given name exists. False otherwise.
        """
            
        r = self.database_renderer(name=name, site=site)
            
#         kwargs = dict(
#             db_user=env.db_root_user,
#             db_password=env.db_root_password,
#             db_host=env.db_host,
#             db_name=env.db_name,
#         )
#         env.update(kwargs)
        
        # Set pgpass file.
#         if env.db_password:
#             self.write_pgpass(verbose=verbose, name=name)
        
#        cmd = ('psql --username={db_user} --no-password -l '\
#            '--host={db_host} --dbname={db_name}'\
#            '| grep {db_name} | wc -l').format(**env)

        ret = None
        with settings(warn_only=True):
            ret = r.run('psql --username={db_user} --host={db_host} -l '\
            '| grep {db_name} | wc -l')
            if ret is not None:
                if 'password authentication failed' in ret:
                    ret = False
                else:
                    ret = int(ret) >= 1
              
        if ret is not None:
            print('%s database on site %s %s exist' % (name, self.genv.SITE, 'DOES' if ret else 'DOES NOT'))
            return ret

    @task
    def execute(self, sql, name='default', site=None, **kwargs):
        r = self.database_renderer(name=name, site=site)
        r.env.sql = sql
        r.run('psql --user={postgres_user} --no-password --command="{sql}"')
        
    @task
    def create(self, name='default', site=None, **kargs):
        
        r = self.database_renderer(name=name, site=site)
        
        # Create role/user.
        with settings(warn_only=True):
            r.pc('Creating user...')
            r.run('psql --user={postgres_user} --no-password --command="CREATE USER {db_user} WITH PASSWORD \'{db_password}\';"')
        
        r.pc('Creating database...')
        r.run('psql --user={postgres_user} --no-password --command="CREATE DATABASE {db_name} WITH OWNER={db_user} ENCODING=\'{encoding}\'"')
        
        with settings(warn_only=True):
            r.pc('Enabling plpgsql on database...')
            r.run('createlang -U postgres plpgsql {db_name}')

    @task
    @runs_once
    def load(self, dump_fn='', prep_only=0, force_upload=0, from_local=0, name=None, site=None, dest_dir=None):
        """
        Restores a database snapshot onto the target database server.
        
        If prep_only=1, commands for preparing the load will be generated,
        but not the command to finally load the snapshot.
        """
        
        r = self.database_renderer(name=name, site=site)
        
        # Render the snapshot filename.
        r.env.dump_fn = self.get_default_db_fn(fn_template=dump_fn, dest_dir=dest_dir)
        
        from_local = int(from_local)
        
        prep_only = int(prep_only)
        
        missing_local_dump_error = r.format(
            "Database dump file {dump_fn} does not exist."
        )
        
        # Copy snapshot file to target.
        if r.genv.is_local:
            r.env.remote_dump_fn = dump_fn
        else:
            r.env.remote_dump_fn = '/tmp/' + os.path.split(r.env.dump_fn)[-1]
        
        #print('r.genv.is_local:', r.genv.is_local, r.genv.hosts)
        if not prep_only and not r.genv.is_local:
            if not self.dryrun:
                assert os.path.isfile(r.env.dump_fn), \
                    missing_local_dump_error
            r.pc('Uploading PostgreSQL database snapshot...')
#                 r.put(
#                     local_path=r.env.dump_fn,
#                     remote_path=r.env.remote_dump_fn)
            r.local('rsync -rvz --progress --no-p --no-g '
                '--rsh "ssh -o StrictHostKeyChecking=no -i {key_filename}" '
                '{dump_fn} {user}@{host_string}:{remote_dump_fn}')
        
        if r.genv.is_local and not prep_only and not self.dryrun:
            assert os.path.isfile(r.env.dump_fn), \
                missing_local_dump_error
        
        with settings(warn_only=True):
            r.run('dropdb --user={db_root_username} {db_name}')
                
        r.run('psql --user={db_root_username} -c "CREATE DATABASE {db_name};"')
        
        with settings(warn_only=True):
            
            if r.env.engine == POSTGIS:
                r.run('psql --user={db_root_username} --no-password --dbname={db_name} --command="CREATE EXTENSION postgis;"')
                r.run('psql --user={db_root_username} --no-password --dbname={db_name} --command="CREATE EXTENSION postgis_topology;"')
        
        with settings(warn_only=True):
            r.run('psql --user={db_root_username} -c "REASSIGN OWNED BY {db_user} TO {db_root_username};"')
            
        with settings(warn_only=True):
            r.run('psql --user={db_root_username} -c "DROP OWNED BY {db_user} CASCADE;"')
            
        r.run('psql --user={db_root_username} -c "DROP USER IF EXISTS {db_user}; '
            'CREATE USER {db_user} WITH PASSWORD \'{db_password}\'; '
            'GRANT ALL PRIVILEGES ON DATABASE {db_name} to {db_user};"')
        for createlang in r.env.createlangs:
            r.env.createlang = createlang
            r.run('createlang -U {db_root_username} {createlang} {db_name} || true')
        
        if not prep_only:
            r.run(r.env.load_command)

    @task
    def shell(self, name='default', site=None, **kwargs):
        """
        Opens a SQL shell to the given database, assuming the configured database
        and user supports this feature.
        """
        r = self.database_renderer(name=name, site=site)
        self.write_pgpass(name=name, site=site, root=True)
        
        db_name = kwargs.get('db_name')
        if db_name:
            r.env.db_name = db_name
            r.run('/bin/bash -i -c "psql --username={db_root_username} --host={db_host} --dbname={db_name}"')
        else:
            r.run('/bin/bash -i -c "psql --username={db_root_username} --host={db_host}"')
        
    @task
    def configure_apt_repository(self):
        r = self.local_renderer
        r.sudo("add-apt-repository '{apt_repo}'")
        r.sudo('wget --quiet -O - {apt_key} | apt-key add -')
        r.sudo('apt-get update') 

    @task
    def version(self):
        r = self.local_renderer
        if self.dryrun:
            v = r.env.version_command
        else:
            v = (r.run('echo {version_command}') or '').strip()
        print(v)
        return v

    @task(precursors=['packager', 'user'])
    def configure(self, *args, **kwargs):
        #TODO:set postgres user password?
        #https://help.ubuntu.com/community/PostgreSQL
        #set postgres ident in pg_hba.conf
        #sudo -u postgres psql postgres
        #sudo service postgresql restart
        #sudo -u postgres psql
        #\password postgres
        r = self.local_renderer

        #self.install_packages()

        if r.env.apt_repo_enabled:
            self.configure_apt_repository()

        r.env.pg_version = self.version()# or r.env.default_version
        
#         r.pc('Backing up PostgreSQL configuration files...')
        r.sudo('cp /etc/postgresql/{pg_version}/main/postgresql.conf /etc/postgresql/{pg_version}/main/postgresql.conf.$(date +%Y%m%d%H%M).bak')
        r.sudo('cp /etc/postgresql/{pg_version}/main/pg_hba.conf /etc/postgresql/{pg_version}/main/pg_hba.conf.$(date +%Y%m%d%H%M).bak')
        
        r.pc('Allowing remote connections...')
        fn = self.render_to_file('postgresql/pg_hba.template.conf')
        r.put(
            local_path=fn,
            remote_path='/etc/postgresql/{pg_version}/main/pg_hba.conf',
            use_sudo=True,
        )
        
        r.pc('Enabling auto-vacuuming...')
        r.sed(  
            filename='/etc/postgresql/{pg_version}/main/postgresql.conf',
            before='#autovacuum = on',
            after='autovacuum = on',
            backup='',
            use_sudo=True,
        )
        r.sed(
            filename='/etc/postgresql/{pg_version}/main/postgresql.conf',
            before='#track_counts = on',
            after='track_counts = on',
            backup='',
            use_sudo=True,
        )
        
        # Set UTF-8 as the default database encoding.
        #TODO:fix? throws error code?
#        sudo_or_dryrun('psql --user=postgres --no-password --command="'
#            'UPDATE pg_database SET datistemplate = FALSE WHERE datname = \'template1\';'
#            'DROP DATABASE template1;'
#            'CREATE DATABASE template1 WITH TEMPLATE = template0 ENCODING = \'UNICODE\';'
#            'UPDATE pg_database SET datistemplate = TRUE WHERE datname = \'template1\';'
#            '\c template1\n'
#            'VACUUM FREEZE;'
#            'UPDATE pg_database SET datallowconn = FALSE WHERE datname = \'template1\';"')

        r.sudo('service postgresql restart')
    
class PostgreSQLClientSatchel(Satchel):

    name = 'postgresqlclient'

    def set_defaults(self):
        self.env.force_version = None
        
    @property
    def packager_system_packages(self):
        return {
            FEDORA: ['postgresql-client'],
#             UBUNTU: [
#                 'postgresql-client-%s' % self.env.default_version,
#                 #'python-psycopg2',#install from pip instead
#                 #'postgresql-server-dev-9.3',
#             ],
            (UBUNTU, '12.04'): [
                'postgresql-client-9.1',
                #'python-psycopg2',#install from pip instead
                #'postgresql-server-dev-9.1',
            ],
            (UBUNTU, '14.04'): [
                'postgresql-client-9.3',
                #'python-psycopg2',#install from pip instead
                #'postgresql-server-dev-9.3',
            ],
            (UBUNTU, '16.04'): [
                'postgresql-client-9.5',
                #'python-psycopg2',#install from pip instead
                #'postgresql-server-dev-9.3',
            ],
        }
        
    @task(precursors=['packager'])
    def configure(self, *args, **kwargs):
        pass

postgresql = PostgreSQLSatchel()
PostgreSQLClientSatchel()

write_postgres_pgpass = postgresql.write_pgpass



import os

from fabric.api import settings

from burlap.constants import *
from burlap import ServiceSatchel
from burlap.decorators import task

class BuildBotSatchel(ServiceSatchel):
    
    name = 'buildbot'
    
    #post_deploy_command = 'reload'
    
    required_system_packages = {
        UBUNTU: ['git'],
        (UBUNTU, '14.04'): ['git'],
        (UBUNTU, '16.04'): ['git'],
    }
    
    def set_defaults(self):
        
        self.env.project_dir = '/usr/local/myproject'
        self.env.virtualenv_dir = '/usr/local/myproject/.env'
        
        # Must match the main user, or otherwise we get rsync errors.
        self.env.user = 'ubuntu'
        self.env.group = 'ubuntu'
        
        self.env.bb_user = 'buildbot'
        self.env.bb_group = 'buildbot'
        
        self.env.manhole_user = 'admin'
        self.env.manhole_port = 1234
        
        self.env.perms = '777'
        
        self.env.cron_path = '/etc/cron.d/buildbot_boot'
        self.env.cron_user = 'root'
        self.env.cron_group = 'root'
        self.env.cron_perms = '600'
        
        self.env.extra_deploy_paths = ['buildbot/slave/info/', 'buildbot/slave/buildbot.tac']
        
        self.env.check_ok = True
        self.env.check_ok_paths = {} # {branch: (url, text)}
        
        self.env.rsync_paths = [] # [[from_path, to_path, user, group]]
        
        self.env.enable_apache_site = True
        
        self.env.requirements = 'pip-requirements.txt'
        
        self.env.service_commands = {
#             START:{
#                 UBUNTU: 'service apache2 start',
#             },
#             STOP:{
#                 UBUNTU: 'service apache2 stop',
#             },
#             DISABLE:{
#                 (UBUNTU, '14.04'): 'update-rc.d -f apache2 remove',
#             },
#             ENABLE:{
#                 (UBUNTU, '14.04'): 'update-rc.d apache2 defaults',
#             },
#             RELOAD:{
#                 UBUNTU: 'service apache2 reload',
#             },
#             RESTART:{
#                 # Note, the sleep 5 is necessary because the stop/start appears to
#                 # happen in the background but gets aborted if Fabric exits before
#                 # it completes.
#                 UBUNTU: 'service apache2 restart; sleep 3',
#             },
        }
    
    @task
    def restart(self):
        self.set_permissions()
        self.restart_master(ignore_errors=True)
        self.restart_slave(ignore_errors=True)
    
    @task
    def restart_master(self, ignore_errors=None):
        ignore_errors = self.ignore_errors if ignore_errors is None else ignore_errors
        r = self.local_renderer
        s = {'warn_only':True} if ignore_errors else {}
        with settings(**s):
            r.run(
                'sudo -u {bb_user} bash -c "cd {project_dir}/src/buildbot; '
                '{virtualenv_dir}/bin/buildbot restart master"')
        
    @task
    def restart_slave(self, ignore_errors=None):
        ignore_errors = self.ignore_errors if ignore_errors is None else ignore_errors
        r = self.local_renderer
        s = {'warn_only':True} if ignore_errors else {}
        with settings(**s):
            r.run(
                'sudo -u {bb_user} bash -c "cd {project_dir}/src/buildbot; '
                '{virtualenv_dir}/bin/buildslave restart slave"')
    
    @task
    def start(self):
        r = self.local_renderer
        s = {'warn_only':True} if self.ignore_errors else {}
        with settings(**s):
            r.run(
                'sudo -u {bb_user} bash -c "cd {project_dir}/src/buildbot; '
                '{virtualenv_dir}/bin/buildbot start master"')
    
    @task
    def stop(self):
        r = self.local_renderer
        s = {'warn_only':True} if self.ignore_errors else {}
        with settings(**s):
            r.run(
                'sudo -u {bb_user} bash -c "cd {project_dir}/src/buildbot; '
                '{virtualenv_dir}/bin/buildbot stop master"')
            r.run(
                'sudo -u {bb_user} bash -c "cd {project_dir}/src/buildbot; '
                '{virtualenv_dir}/bin/buildbot stop slave"')
    
    @task
    def reload(self):
        r = self.local_renderer
        s = {'warn_only':True} if self.ignore_errors else {}
        with settings(**s):
            r.run(
                'sudo -u {bb_user} bash -c "'
                'cd {project_dir}/src/buildbot; '
                '{virtualenv_dir}/bin/buildbot reconfig master"')
    
    @task
    def set_permissions(self):
        r = self.local_renderer
        r.sudo('chown -R {bb_user}:{bb_group} {project_dir}')
        #r.sudo('chown -R {bb_user}:{bb_group} {project_dir}/src/buildbot/master')
        #r.sudo('chown -R {bb_user}:{bb_group} {project_dir}/src/buildbot/slave')
        r.sudo('chmod -R {perms} {project_dir}')
    
    @task
    def rsync_paths(self):
        r = self.local_renderer
        for from_path, to_path, to_user, to_group, to_perms in r.env.rsync_paths:
            assert os.path.isfile(from_path)
            r.env.from_path = from_path
            r.env.to_path = to_path
            r.env.to_user = to_user
            r.env.to_group = to_group
            r.env.to_path_fq = os.path.join(to_path, os.path.split(from_path)[-1])
            r.env.to_perms = to_perms
#             r.local('rsync '
#                 '--verbose --compress '
#                 '--rsh "ssh -t -o StrictHostKeyChecking=no -i {key_filename}" '
#                 '{from_path} {user}@{host_string}:{to_path}')
            r.put(local_path=r.env.from_path, remote_path=r.env.to_path, use_sudo=True)
            r.sudo('chown {to_user}:{to_group} {to_path_fq}')
            r.sudo('chmod {to_perms} {to_path_fq}')
    
    @task
    def setup_dir(self, clean=0):
        pip = self.get_satchel('pip')
        clean = int(clean)
        r = self.local_renderer
        if clean:
            r.sudo('rm -Rf {virtualenv_dir} || true')
        r.sudo('mkdir -p {project_dir}')
        pip.update_install(
            requirements=self.env.requirements,
            virtualenv_dir=self.env.virtualenv_dir,
            user=self.env.user,
            group=self.env.group,
            perms=self.env.perms,
        )
        #r.sudo('{virtualenv_dir}/bin/pip install -U pip')
        r.sudo('chown -R {user}:{group} {project_dir}')
        r.sudo('chmod -R {perms} {project_dir}')
    
    @task
    def install_cron(self):
        r = self.local_renderer
        r.sudo(
            'printf \'SHELL=/bin/bash\\nPATH=/usr/local/sbin:/usr/local/bin:/sbin:/bin:/usr/sbin:/usr/bin\\n@reboot '
            'buildbot bash -c "cd {project_dir}/src/buildbot; '
            '{project_dir}/.env/bin/buildbot start master; '
            '{project_dir}/.env/bin/buildslave start slave"\\n\' > {cron_path}')
        r.sudo('chown {cron_user}:{cron_group} {cron_path}')
        # Must be 600, otherwise gives INSECURE MODE error.
        # http://unix.stackexchange.com/questions/91202/cron-does-not-print-to-syslog
        r.sudo('chmod {cron_perms} {cron_path}')
        r.sudo('service cron restart')
    
    @task
    def deploy_code(self):
        r = self.local_renderer
        
        r.sudo('chown -R {user}:{group} {project_dir}')
        r.sudo('chmod -R {perms} {project_dir}')
        
        r.local('rsync '
            '--recursive --verbose --perms --times --links '
            '--compress --copy-links '
            '--exclude=*.pyc --exclude=gitpoller-workdir --exclude=*.log --exclude=twistd.pid '
            '--exclude=*.sqlite '
            '--exclude=build '
            '--exclude=slave '
            '--exclude=*_runtests '
            '--delete --delete-before '
            '--rsh "ssh -t -o StrictHostKeyChecking=no -i {key_filename}" '
            'src {user}@{host_string}:{project_dir}')
        
        for path in self.env.extra_deploy_paths:
            if path.endswith('/'):
                # Directory.
                r.env.tmp_path = path[:-1]
                r.env.tmp_remote_path = os.path.split(path[:-1])[0]
                r.sudo('mkdir -p {project_dir}/src/{tmp_remote_path}')
            else:
                # File.
                r.env.tmp_path = path
                r.env.tmp_remote_path = os.path.split(path)[0]

            r.pc('Fixing permissions...')
            r.sudo('chown -R {bb_user}:{bb_group} {project_dir}')
            r.sudo('chmod -R {perms} {project_dir}')
            r.local('rsync '
                '--recursive --verbose --times --omit-dir-times --links '
                '--compress --copy-links '
                '--delete --delete-before '
                '--rsh "ssh -t -o StrictHostKeyChecking=no -i {key_filename}" '
                'src/{tmp_path} {user}@{host_string}:{project_dir}/src/{tmp_remote_path}')
        
        self.rsync_paths()
        
        self.set_permissions()
    
    @task
    def view_log(self, name='master', section='twistd'):
        r = self.local_renderer
        r.env.name = name
        assert section in ('twistd', 'http')
        r.env.section = section
        r.run('tail -f --lines=100 {project_dir}/src/buildbot/{name}/{section}.log')
    
    @task
    def manhole(self):
        r = self.local_renderer
        r.env.host_string = '127.0.0.1'#self.genv.host_string
        r.run('ssh -p{manhole_port} {manhole_user}@{host_string}')
    
    @task
    def check_ok(self):
        """
        Ensures all tests have passed for this branch.
        
        This should be called before deployment, to prevent accidental deployment of code
        that hasn't passed automated testing.
        """
        import requests
        
        if not self.env.check_ok:
            return
        
        # Find current git branch.
        branch_name = self._local('git rev-parse --abbrev-ref HEAD', capture=True).strip()
        
        check_ok_paths = self.env.check_ok_paths or {}
        
        if branch_name in check_ok_paths:
            check = check_ok_paths[branch_name]
            if 'username' in check:
                auth = (check['username'], check['password'])
            else:
                auth = None
            ret = requests.get(check['url'], auth=auth)
            passed = check['text'] in ret.content
            assert passed, 'Check failed: %s' % check['url']
    
    @task
    def setup_user(self):
        user = self.get_satchel('user')
        group = self.get_satchel('group')
        user.create(self.env.user, self.env.bb_group)
        group.create(self.env.bb_group)
        user.create(self.env.bb_user, self.env.bb_group)
    
    def deploy_pre_run(self):
        self.check_ok()
        
    @task(precursors=['packager', 'user', 'apache'])
    def configure(self):
        packager = self.get_satchel('packager')
#         umv = self.get_satchel('ubuntumultiverse')
        
        packager.configure()
        
        if self.env.enable_apache_site:
            apache = self.get_satchel('apache')
            apache.enable_mod('proxy_http')
        
        self.setup_user()
        
        #umv.configure()
        
        # Initialize base project directory and        
        # Setup Python virtual environment.
        self.setup_dir()
        
        # Copy up our code.
        self.deploy_code()
        
        self.install_cron()
        
buildbot = BuildBotSatchel()



# 
# import pytest
# 
# 
# pytestmark = pytest.mark.network
# 
# 
# @pytest.fixture(scope='module')
# def postgres_server():
#     from burlap.require.postgresql import server
#     server()
# 
# 
# @pytest.yield_fixture(scope='module')
# def postgres_user():
#     from burlap.postgresql import drop_user
#     from burlap.require.postgresql import user
#     name = 'pguser'
#     user(name, password='s3cr3t')
#     yield name
#     drop_user(name)
# 
# 
# def test_create_and_drop_user(postgres_server):
#     from burlap.postgresql import create_user, drop_user, user_exists
# 
#     create_user('alice', password='1234')
#     assert user_exists('alice')
# 
#     drop_user('alice')
#     assert not user_exists('alice')
# 
# 
# def test_require_user(postgres_server):
#     from burlap.postgresql import user_exists, drop_user
#     from burlap.require.postgresql import user
# 
#     user('bob', password='foo')
# 
#     assert user_exists('bob')
# 
#     drop_user('bob')
# 
# 
# def test_require_database(postgres_server, postgres_user):
#     from burlap.postgresql import database_exists, drop_database
#     from burlap.require.postgresql import database
# 
#     database('pgdb', postgres_user)
# 
#     assert database_exists('pgdb')
# 
#     drop_database('pgdb')



# from pipes import quote
# from textwrap import dedent
# import posixpath
# 
# import pytest
# 
# from fabric.api import quiet, run, shell_env, sudo
# 
# from burlap.files import is_link
# from burlap.system import distrib_family
# 
# def test_apache():
#     from burlap.project import Project
#     project = Project(name='test_apache', base_dir='/tmp')
#     project.init()
#     project.add_role('prod')
#     project.roles.prod.add_service('apache')
#     project.deploy()



import unittest

from fabric.contrib.files import exists

from burlap.selenium import selenium

class SeleniumTests(unittest.TestCase):

    def test_selenium(self):

        print('selenium.geckodriver_path:', selenium.geckodriver_path)

        selenium.env.enabled = True
        selenium.configure()

        assert exists(selenium.geckodriver_path)

        selenium.env.enabled = False
        selenium.configure()

        assert not exists(selenium.geckodriver_path)



from __future__ import print_function

import sys
import re
import traceback
from pprint import pprint

from fabric.api import (
    env,
)

from burlap import common
from burlap.decorators import task_or_dryrun

if 'jira_server' not in env:
    env.jira_server = None
    env.jira_basic_auth_username = None
    env.jira_basic_auth_password = None
    env.jira_update_from_git = False
    env.jira_ticket_update_message_template = 'This has been deployed to %(role)s.'
    
    # A map of status->transition to follow when making deployments.
    env.jira_deploy_workflow = {}
    
    # A map of the new-status->new-assignee to auto-assign.
    env.jira_assignee_by_status = {}
    
    # The regex used to search the commit messages for a ticket number.
    env.jira_ticket_pattern = None

@task_or_dryrun
def get_tickets_between_commits(a, b):
    from burlap.git import gittracker
    
    get_logs_between_commits = gittracker.get_logs_between_commits
    
    tickets = []
    if env.jira_ticket_pattern:
        verbose = common.get_verbose()
        ret = get_logs_between_commits(a, b)
        pattern = re.compile(env.jira_ticket_pattern, flags=re.I)
        if verbose:
            print('pattern:', env.jira_ticket_pattern)
        tickets.extend(pattern.findall(ret))
    if verbose:
        print(tickets)
    return set(_.strip().upper() for _ in tickets)

@task_or_dryrun
def update_tickets_from_git(last=None, current=None):
    """
    Run during a deployment.
    Looks at all commits between now and the last deployment.
    Finds all ticket numbers and updates their status in Jira.
    """
    from jira import JIRA, JIRAError
    from burlap.deploy import get_last_current_diffs
    from burlap.git import gittracker
    
    get_current_commit = gittracker.get_current_commit
    GITTRACKER = gittracker.name.upper()
    
    dryrun = common.get_dryrun()
    verbose = common.get_verbose()
    
    # Ensure this is only run once per role.
    if env.host_string != env.hosts[-1]:
        return
    
    if not env.jira_update_from_git:
        return
    
    if not env.jira_ticket_pattern:
        return
    
    if not env.jira_basic_auth_username or not env.jira_basic_auth_password:
        return
    
    # During a deployment, we should be given these, but for testing,
    # lookup the diffs dynamically.
    if not last or not current:
        last, current = get_last_current_diffs(GITTRACKER)
    
    if verbose:
        print('-'*80)
        print('last.keys:', last.keys())
        print('-'*80)
        print('current.keys:', current.keys())
    
    try:
        last_commit = last['GITTRACKER']['current_commit']
    except KeyError:
        return
    current_commit = current['GITTRACKER']['current_commit']
    
    # Find all tickets deployed between last deployment and now.
    tickets = get_tickets_between_commits(current_commit, last_commit)
    if verbose:
        print('tickets:', tickets)
    
    # Update all tickets in Jira.
    jira = JIRA({
        'server': env.jira_server
    }, basic_auth=(env.jira_basic_auth_username, env.jira_basic_auth_password))
    for ticket in tickets:
        
        # Mention this Jira updated.
        comment = env.jira_ticket_update_message_template % dict(role=env.ROLE.lower())
        print('Commenting on ticket %s: %s' % (ticket, comment))
        if not dryrun:
            jira.add_comment(ticket, comment) 
        
        # Update ticket status.
        recheck = False
        while 1:
            print('Looking up jira ticket %s...' % ticket)
            issue = jira.issue(ticket)
            print('Ticket %s retrieved.' % ticket)
            transition_to_id = dict((t['name'], t['id']) for t in jira.transitions(issue))
            print('%i allowable transitions found:')
            pprint(transition_to_id)
            print('issue.fields.status.id:', issue.fields.status.id)
            print('issue.fields.status.name:', issue.fields.status.name)
            jira_status_id = issue.fields.status.name.title()
            print('jira_status_id:', jira_status_id)
            next_transition_name = env.jira_deploy_workflow.get(jira_status_id)
            print('next_transition_name:', next_transition_name)
            next_transition_id = transition_to_id.get(next_transition_name)
            print('next_transition_id:', next_transition_id)
            if next_transition_name:
                new_fields = {}
                new_assignee = env.jira_assignee_by_status.get(
                    #issue.fields.status.name.title(),
                    next_transition_name,
                    issue.fields.assignee.name,
                )
                if new_assignee == 'reporter':
                    new_assignee = issue.fields.reporter.name
#                 print('new_assignee:', new_assignee)
                    
                print('Updating ticket %s to status %s (%s) and assigning it to %s.' \
                    % (ticket, next_transition_name, next_transition_id, new_assignee))
                if not dryrun:
                    try:
                        jira.transition_issue(
                            issue,
                            next_transition_id,
                        )
                        recheck = True
                    except AttributeError as e:
                        print('Unable to transition ticket %s to %s: %s' \
                            % (ticket, next_transition_name, e), file=sys.stderr)
                        traceback.print_exc()
                    
                    # Note assignment should happen after transition, since the assignment may
                    # effect remove transitions that we need.
                    try:
                        if new_assignee:
                            print('Assigning ticket %s to %s.' % (ticket, new_assignee))
                            jira.assign_issue(issue, new_assignee)
                        else:
                            print('No new assignee found.')
                    except JIRAError as e:
                        print('Unable to reassign ticket %s to %s: %s' \
                            % (ticket, new_assignee, e), file=sys.stderr)
            else:
                recheck = False
                print('No transitions found for ticket %s currently in status "%s".' \
                    % (ticket, issue.fields.status.name))
                    
            if not recheck:
                break



 
from burlap.files import file # pylint: disable=redefined-builtin
from burlap.user import user

is_dir = file.is_dir

def test_create_user():
    try:
        user.create('user1', create_home=False, password=False)
        assert user.exists('user1')
        assert not is_dir('/home/user1')
    finally:
        user.sudo('userdel -r user1')



from mock import patch

import pytest


def test_unsupported_system():

    from burlap.system import UnsupportedFamily

    with pytest.raises(UnsupportedFamily) as excinfo:

        with patch('burlap.system.distrib_id') as mock_distrib_id:
            mock_distrib_id.return_value = 'foo'

            raise UnsupportedFamily(supported=['debian', 'redhat'])

    exception_msg = str(excinfo.value)
    assert exception_msg == "Unsupported family other (foo). Supported families: debian, redhat"



"""
Convenience decorators for use in fabfiles.
"""
from __future__ import with_statement

#from Crypto import Random

#from fabric import tasks
from fabric.api import runs_once as _runs_once

#from .context_managers import settings
from burlap.tasks import WrappedCallableTask

def task_or_dryrun(*args, **kwargs):
    """
    Decorator declaring the wrapped function to be a new-style task.

    May be invoked as a simple, argument-less decorator (i.e. ``@task``) or
    with arguments customizing its behavior (e.g. ``@task(alias='myalias')``).

    Please see the :ref:`new-style task <task-decorator>` documentation for
    details on how to use this decorator.

    .. versionchanged:: 1.2
        Added the ``alias``, ``aliases``, ``task_class`` and ``default``
        keyword arguments. See :ref:`task-decorator-arguments` for details.
    .. versionchanged:: 1.5
        Added the ``name`` keyword argument.

    .. seealso:: `~fabric.docs.unwrap_tasks`, `~fabric.tasks.WrappedCallableTask`
    """
    invoked = bool(not args or kwargs)
    task_class = kwargs.pop("task_class", WrappedCallableTask)
#     if invoked:
#         func, args = args[0], ()
#     else:
    func, args = args[0], ()

    def wrapper(func):
        return task_class(func, *args, **kwargs)
    wrapper.is_task_or_dryrun = True
    wrapper.wrapped = func

    return wrapper if invoked else wrapper(func)

def _task(meth):
    meth.is_task = True
    return meth

def task(*args, **kwargs):
    """
    Decorator for registering a satchel method as a Fabric task.
    
    Can be used like:
    
        @task
        def my_method(self):
            ...
            
        @task(precursors=['other_satchel'])
        def my_method(self):
            ...

    """
    precursors = kwargs.pop('precursors', None)
    post_callback = kwargs.pop('post_callback', False)
    if args and callable(args[0]):
        # direct decoration, @task
        return _task(*args)
    else:
        # callable decoration, @task(precursors=['satchel'])
        def wrapper(meth):
            if precursors:
                meth.deploy_before = list(precursors)
            if post_callback:
                #from burlap.common import post_callbacks
                #post_callbacks.append(meth)
                meth.is_post_callback = True
            return _task(meth)
        return wrapper

def runs_once(meth):
    """
    A wrapper around Fabric's runs_once() to support our dryrun feature.
    """
    from burlap.common import get_dryrun
    if get_dryrun():
        pass
    else:
        _runs_once(meth)
    return meth
